package be.iminds.ilabt.jfed.preferences;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.PreferencesPuttySshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.PreferencesSshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.TerminalFinder;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.OSDetector;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * User: twalcari
 * Date: 11/18/13
 * Time: 6:03 PM
 */
public class JFedGuiPreferences extends JFedCorePreferences {
    private final static Logger LOG = LoggerFactory.getLogger(JFedGuiPreferences.class);

    private static final String DEFAULT_PUTTY_DIR = "PuTTY";
    private static final String[] DEFAULT_PUTTY_LOCATIONS =
            new String[]{System.getenv("ProgramFiles") + File.separator + DEFAULT_PUTTY_DIR,
                    System.getenv("ProgramFiles(x86)") + File.separator + DEFAULT_PUTTY_DIR, "C:\\PuTTY"};


    public JFedGuiPreferences() {
        super();

        config.setAutoSave(true);
    }

    /* handles all errors, by returning null */
    private static String readOpenSshVersion(String opt) {
        try {
            ArrayList<String> command = new ArrayList<>();
            command.add("ssh");
            command.add(opt);
            Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));

            BufferedReader input = new BufferedReader(new InputStreamReader(p.getErrorStream()));
            p.getOutputStream().close();
            String versionLine = input.readLine();
            input.close();

            try {
                p.waitFor();
            } catch (InterruptedException e) {
                //ignore
            }

            return versionLine;
        } catch (AssertionError e) {
            LOG.error("AssertionError trying to call \"ssh "+opt+"\": " + e.getMessage(), e);
            return null;
        } catch (Exception e) {
            LOG.error("Exception trying to call \"ssh "+opt+"\": " + e.getMessage(), e);
            return null;
        }
    }

    /** returns null if unknown */
    private static Boolean checkIfNewVersion(String versionLine) {
        if (versionLine == null) return null;
        Pattern patternVersion = Pattern.compile("OpenSSH.?([0-9]+)\\.([0-9]+).*");
        Matcher matcherVersion = patternVersion.matcher(versionLine);
        if (versionLine != null && matcherVersion.find()) {
            int bigVersion = Integer.parseInt(matcherVersion.group(1));
            int smallVersion = Integer.parseInt(matcherVersion.group(2));
            LOG.info("OpenSSH version is: " + bigVersion + "." + smallVersion);
            if (bigVersion > 5)
                return true;
            if (bigVersion == 5 && smallVersion >= 4)
                return true;
            return false;
        } else
            return null;
    }


    protected void fillinBlanks(PropertiesConfiguration config) {
        super.fillinBlanks(config);

        if (OSDetector.getUnixLikeOsFilter().includedOs(OSDetector.os)) {
            if (!config.containsKey(GuiPreferenceKey.PREF_COMMAND_UNIX_TERMINAL.key))
                config.setProperty(GuiPreferenceKey.PREF_COMMAND_UNIX_TERMINAL.key, findUnixTerminalCommand());

            boolean autoSshCommand = getBoolean(GuiPreferenceKey.PREF_COMMAND_SSH_AUTO, true);

            if (autoSshCommand || !config.containsKey(GuiPreferenceKey.PREF_COMMAND_SSH.key))
                config.setProperty(GuiPreferenceKey.PREF_COMMAND_SSH.key, "ssh -A -X -i '$i' $u@$h -oPort=$p");

            final String defaultUnixSshCommandWithoutNewOpenSsh =
                    "ssh -A -X -i '$i' $u@$h -oPort=$p -oProxyCommand=\"ssh -e none -i '$I' -oPort=$P $U@$H nc -w 5 %h %p\"";
            final String defaultUnixshCommandWithNewOpenSsh =
                    "ssh -A -X -i '$i' $u@$h -oPort=$p -oProxyCommand=\"ssh -i '$I' -oPort=$P $U@$H -W %h:%p\"";

            if (autoSshCommand ||
                    (!config.containsKey(GuiPreferenceKey.PREF_COMMAND_SSH_PROXY.key)) ||
                    config.getProperty(GuiPreferenceKey.PREF_COMMAND_SSH_PROXY.key).equals(defaultUnixSshCommandWithoutNewOpenSsh)) {
                /**
                 * OpenSSH version >= 5.4 supports the -W host:port option, which does the netcat trick, without netcat.
                 *
                 *      -W host:port
                 *            Requests that standard input and output on the client be
                 *            forwarded to host on port over the secure channel.  Implies -N,
                 *            -T, ExitOnForwardFailure and ClearAllForwardings.  Works with
                 *            Protocol version 2 only.
                 * */
                boolean newOpenSsh = false;
                //we are on unix, so we call openssh:
                //   ssh -v
                // should return something starting with: OpenSSH_5.3p1 Debian-3ubuntu7.1, OpenSSL 0.9.8k 25 M...
                //  On newer version, it's:   ssh -V
                // so we have to try both


                String versionLine = readOpenSshVersion("-V");
                LOG.info("Read \"ssh -V\" output: \"" + versionLine + "\"");
                Boolean check1 = checkIfNewVersion(versionLine);
                if (check1 != null) {
                    newOpenSsh = check1;
                } else {
                    versionLine = readOpenSshVersion("-v");
                    LOG.info("Read \"ssh -v\" output: \"" + versionLine + "\"");
                    Boolean check2 = checkIfNewVersion(versionLine);
                    if (check2 != null) {
                        newOpenSsh = check2;
                    } else {
                        LOG.info("OpenSSH version could not be determined");
                    }
                }

                LOG.debug("newOpenSsh=" + newOpenSsh + "");

                if (newOpenSsh)
                    config.setProperty(GuiPreferenceKey.PREF_COMMAND_SSH_PROXY.key, defaultUnixshCommandWithNewOpenSsh);
                else
                    config.setProperty(GuiPreferenceKey.PREF_COMMAND_SSH_PROXY.key, defaultUnixSshCommandWithoutNewOpenSsh);
            }
        }


        if (OSDetector.findOS() == OSDetector.OS.WIN) {
            //try to find settings for PuTTY
            if (!config.containsKey(GuiPreferenceKey.PREF_PUTTY_DIRECTORY.key)) {
                //try to detect the location of PuTTY
                String puttyDir = JFedGuiPreferences.findPuTTYDirectory();

                boolean enablePageant = true;

                if (puttyDir != null) {
                    config.setProperty(GuiPreferenceKey.PREF_PUTTY_DIRECTORY.key, puttyDir);
                    enablePageant = JFedGuiPreferences.pageantExists(new File(puttyDir));
                } else{
                    //disable pageant, as we didn't find a suitable PuTTY-dir
                    enablePageant = false;
                }

                if (!config.containsKey(GuiPreferenceKey.PREF_SSHAGENT_USE.key))
                    config.setProperty(GuiPreferenceKey.PREF_SSHAGENT_USE.key, enablePageant);

            }


        }

        if (!config.containsKey(GuiPreferenceKey.PREF_FIRSTRUN_VERSION.key))
            config.setProperty(GuiPreferenceKey.PREF_FIRSTRUN_VERSION.key, 0);

        if (!config.containsKey(GuiPreferenceKey.PREF_SSHAGENT_USE.key))
            config.setProperty(GuiPreferenceKey.PREF_SSHAGENT_USE.key, true);

        if (!config.containsKey(GuiPreferenceKey.PREF_SSH_OVERRIDE.key))
            config.setProperty(GuiPreferenceKey.PREF_SSH_OVERRIDE.key, false);

        if (!config.containsKey(GuiPreferenceKey.PREF_SSHPROXY_OVERRIDE.key))
            config.setProperty(GuiPreferenceKey.PREF_SSHPROXY_OVERRIDE.key, false);

        if (!config.containsKey(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED.key))
            config.setProperty(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED.key, ProxyPreferencesManager.NEVER_USE_PROXY);

        if (!config.containsKey(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH.key))
            config.setProperty(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH.key, ProxyPreferencesManager.NEVER_USE_PROXY);

        if (!config.containsKey(GuiPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES.key))
            config.setProperty(GuiPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES.key, true);

        if (!config.containsKey(GuiPreferenceKey.PREF_TERMINAL_DEBUG.key))
            config.setProperty(GuiPreferenceKey.PREF_TERMINAL_DEBUG.key, true);

        if (!config.containsKey(GuiPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED.key))
            config.setProperty(GuiPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED.key, true);

        if (!config.containsKey(GuiPreferenceKey.PREF_EXPIRATION_WARNING_TIME.key))
            config.setProperty(GuiPreferenceKey.PREF_EXPIRATION_WARNING_TIME.key, 60);

        if (!config.containsKey(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED.key))
            config.setProperty(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED.key, true);

        if (!config.containsKey(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME.key))
            config.setProperty(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME.key, 110);
    }

    public void setScs(SfaAuthority auth) {
        if (auth == null) {
            clear(GuiPreferenceKey.PREF_SCS_URN);
            clear(GuiPreferenceKey.PREF_SCS_URL);
            return;
        }
        URL u = auth.getUrl(ServerType.GeniServerRole.SCS, 1);
        if (u == null) throw new RuntimeException("Authority " + auth.getUrnString() + " has no SCS Url");
        setScs(auth.getUrn(), u);
    }

    public void setScs(GeniUrn scsUrn, URL url) {
        if (scsUrn == null && url == null) {
            clear(GuiPreferenceKey.PREF_SCS_URN);
            clear(GuiPreferenceKey.PREF_SCS_URL);
            return;
        }
        if (scsUrn != null)
            setString(GuiPreferenceKey.PREF_SCS_URN, scsUrn.toString());
        else
            clear(GuiPreferenceKey.PREF_SCS_URN);
        setString(GuiPreferenceKey.PREF_SCS_URL, url.toString());
    }

    public void setScs(URL url) {
        if (url == null) {
            clear(GuiPreferenceKey.PREF_SCS_URN);
            clear(GuiPreferenceKey.PREF_SCS_URL);
            return;
        }
        clear(GuiPreferenceKey.PREF_SCS_URN);
        setString(GuiPreferenceKey.PREF_SCS_URL, url.toString());
    }

    public URL getScsUrl() {
        String u = getString(GuiPreferenceKey.PREF_SCS_URL);
        if (u == null) return null;
        try {
            return new URL(u);
        } catch (MalformedURLException e) {
            clear(GuiPreferenceKey.PREF_SCS_URL);
            throw new RuntimeException("Stored SCS URL is not valid URL (and has been removed from config). value=\"" + u + "\"");
        }
    }

    public GeniUrn getScsUrn() {
        String u = getString(GuiPreferenceKey.PREF_SCS_URN);
        if (u == null) return null;
        try {
            return new GeniUrn(u);
        } catch (GeniUrn.GeniUrnParseException e) {
            clear(GuiPreferenceKey.PREF_SCS_URN);
            throw new RuntimeException("Stored SCS URN is not valid URN (and has been removed from config). value=\"" + u + "\"");
        }
    }

    public SfaAuthority getScsAuthority(AuthorityListModel authorityListModel) {
        URL scsUrl = getScsUrl();
        String scsUrnString = getString(GuiPreferenceKey.PREF_SCS_URN);
        GeniUrn scsUrn = scsUrnString == null ? null : GeniUrn.parse(scsUrnString);

        SfaAuthority scsAuth = null;
        if (scsAuth == null && scsUrn != null) {
            scsAuth = authorityListModel.getByUrnExact(scsUrn);
            assert scsAuth == null || scsAuth.getUrls().containsValue(scsUrl);
        }
        if (scsAuth == null && scsUrl != null)
            scsAuth = authorityListModel.getByUrl(scsUrl);
        if (scsAuth == null && scsUrl != null) {
            Map<ServerType, URL> urlMap = new HashMap<>();
            urlMap.put(new ServerType(ServerType.GeniServerRole.SCS, 1), scsUrl);
            try {
                if (scsUrn == null || !scsUrn.getEncodedTopLevelAuthority().equals(scsUrl.getAuthority()))
                    scsUrn = GeniUrn.createGeniUrnFromEncodedParts(scsUrl.getAuthority(), "auth", "scs");
                scsAuth = new SfaAuthority(scsUrn.toString(), "Stitching SCS", null, urlMap, null/*proxies*/, null, "scs");
//                authorityListModel.addAuthority(scsAuth);
//                authorityListModel.fireChange();
            } catch (JFedException e) {
                LOG.error("Something went wrong while creating an SfaAuthority for the SCS: " + e.getMessage(), e);
                throw new RuntimeException("Something went wrong while creating an SfaAuthority for the SCS: " + e.getMessage(), e);
            }
        }
        assert scsAuth == null || scsAuth.getUrl(ServerType.GeniServerRole.SCS, 1) != null;

        return scsAuth;
    }

    public SshKeyInfo getOverriddenSshKeyInfo() {
        if (getBoolean(GuiPreferenceKey.PREF_SSH_OVERRIDE)) {
            if (config.containsKey(GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE.getKey())) {
                PreferencesPuttySshKeyInfo res =
                        new PreferencesPuttySshKeyInfo(this, GuiPreferenceKey.PREF_SSH_OVERRIDE,
                                GuiPreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
                if (!res.isComplete()) return null;
                return res;
            } else {
                PreferencesSshKeyInfo res = new PreferencesSshKeyInfo(this, GuiPreferenceKey.PREF_SSH_OVERRIDE,
                        GuiPreferenceKey.PREF_SSH_PUBLIC_KEY_FILE,
                        GuiPreferenceKey.PREF_SSH_PRIVATE_KEY_FILE,
                        GuiPreferenceKey.PREF_SSH_PUBLIC_KEY,
                        GuiPreferenceKey.PREF_SSH_PRIVATE_KEY);
                if (!res.isComplete()) return null;
                return res;
            }
        } else
            return null;
    }

    public SshKeyInfo getOverriddenProxySshKeyInfo() {
        if (getBoolean(GuiPreferenceKey.PREF_SSHPROXY_OVERRIDE)) {
            if (config.containsKey(GuiPreferenceKey.PREF_SSHPROXY_PUTTY_KEY_FILE.getKey())) {
                PreferencesPuttySshKeyInfo res =
                        new PreferencesPuttySshKeyInfo(this, GuiPreferenceKey.PREF_SSH_OVERRIDE,
                                GuiPreferenceKey.PREF_SSHPROXY_PUTTY_KEY_FILE);
                if (!res.isComplete()) return null;
                return res;
            } else {
                PreferencesSshKeyInfo res = new PreferencesSshKeyInfo(this, GuiPreferenceKey.PREF_SSHPROXY_OVERRIDE,
                        GuiPreferenceKey.PREF_SSHPROXY_PUBLIC_KEY_FILE,
                        GuiPreferenceKey.PREF_SSHPROXY_PRIVATE_KEY_FILE,
                        GuiPreferenceKey.PREF_SSHPROXY_PUBLIC_KEY,
                        GuiPreferenceKey.PREF_SSHPROXY_PRIVATE_KEY);
                if (!res.isComplete()) return null;
                return res;
            }
        } else
            return null;
    }

    public boolean isSshAgentAvailable() {
        if (OSDetector.os == OSDetector.OS.WIN) {
            String puttyDir = getString(GuiPreferenceKey.PREF_PUTTY_DIRECTORY);
            if (puttyDir == null) return false;
            return new File(puttyDir, "pageant.exe").exists();
        } else {
            //linux or mac
            if (OSDetector.os == OSDetector.OS.MAC)
                return false;
            return true;
        }
    }

    public static boolean isValidPuttyDir(File puttyDir) {
        if (puttyDir == null) return false;
//        LOG.debug("isValidPuttyDir("+puttyDir.getPath()+") exists="+puttyDir.exists()+" isDir="+puttyDir.isDirectory());
        if (!puttyDir.exists()) return false;
        if (!puttyDir.isDirectory()) return false;
        if (!new File(puttyDir, "putty.exe").exists()) return false;
//        if (!new File(puttyDir, "puttygen.exe").exists()) return false;
        if (!new File(puttyDir, "plink.exe").exists()) return false;
        return true;
    }



    public static String findPuTTYDirectory() {
        for (String location : DEFAULT_PUTTY_LOCATIONS) {
            File file = new File(location);
            if (isValidPuttyDir(file))
                return location;
        }
        return null;

    }

    private static boolean pageantExists(File puttyDir) {
        return new File(puttyDir, "pageant.exe").exists();
    }


    private String autoFindUnixTerminalCommand;
    private String findUnixTerminalCommand() {
        if (autoFindUnixTerminalCommand == null) {
            if (OSDetector.os == OSDetector.OS.UNIX)
                autoFindUnixTerminalCommand = TerminalFinder.findLinuxTerminalCommand();

            else if (OSDetector.os == OSDetector.OS.MAC)
                autoFindUnixTerminalCommand = "tell application \"Terminal\" to do script \"%\"";
        }
        if (autoFindUnixTerminalCommand != null) {
            LOG.info("Falling back to auto-detected terminal command: " + autoFindUnixTerminalCommand);
        }
        return autoFindUnixTerminalCommand;

    }

    public boolean isExpirationWarningEnabled() {
        return getBoolean(GuiPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED, true);
    }


    public void setExpirationWarningEnabled(boolean enabled) {
        setBoolean(GuiPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED, enabled);
    }


    public int getExpirationWarningTime() {
        return getInt(GuiPreferenceKey.PREF_EXPIRATION_WARNING_TIME, 60);
    }


    public void setExpirationWarningTime(int value) {
        setInt(GuiPreferenceKey.PREF_EXPIRATION_WARNING_TIME, value);
    }


    public boolean isRecoverExpirationWarningEnabled() {
        return getBoolean(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED, true);
    }


    public void setRecoverExpirationWarningEnabled(boolean enabled) {
        setBoolean(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED, enabled);
    }


    public int getRecoverExpirationWarningTime() {
        return getInt(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME, 10);
    }


    public void setRecoverExpirationWarningTime(int value) {
        setInt(GuiPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME, value);
    }
}
