package be.iminds.ilabt.jfed.connectivity_tester;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.BasicSshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.stitching.VlanRange;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;

import javax.inject.Inject;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * User: twalcari
 * Date: 1/7/14
 * Time: 2:28 PM
 */
public class ProxyTestGenerator implements TestsGenerator {

    private static final String WALL2_URN = "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm";
    private final GeniUserProvider geniUserProvider;
    private final AuthorityListModel authorityListModel;
    private final ConnectivityTestFactory connectivityTestFactory;
    private final JFedGuiPreferences jFedPreferences;

    @Inject
    public ProxyTestGenerator(GeniUserProvider geniUserProvider, AuthorityListModel authorityListModel, ConnectivityTestFactory connectivityTestFactory, JFedGuiPreferences jFedPreferences) {
        this.geniUserProvider = geniUserProvider;
        this.authorityListModel = authorityListModel;
        this.connectivityTestFactory = connectivityTestFactory;
        this.jFedPreferences = jFedPreferences;
    }


    @Override
    public Collection<ConnectivityTest> generateTests() {

        List<ConnectivityTest> tests = new ArrayList<>();

        List<SfaAuthority> authorities = authorityListModel.getAuthorities();

        for (SfaAuthority auth : authorities) {
            for (SfaAuthority.ProxyInfo proxyInfo : auth.getProxies()) {
                VlanRange portRange = new VlanRange(proxyInfo.getPortRange());
                Integer port = portRange.getFirst();

                //add test of proxy usage if there is a user logged in and this is his authority
                if (port != null && geniUserProvider != null &&
                        geniUserProvider.isUserLoggedIn() &&
                        geniUserProvider.getLoggedInGeniUser().getUserAuthority().equals(auth)) {
                    GeniUser user = geniUserProvider.getLoggedInGeniUser();
                    assert user != null;

                    String username = user.getUserUrn().getEncodedResourceName();

                    SshKeyInfo userSshKeyInfo = new BasicSshKeyInfo(
                            user.getPublicKey(),
                            user.getPrivateKey());

                    JFedConnection.SshProxyInfo conProxyInfo = new JFedConnection.SshProxyInfo(
                            proxyInfo.getHostname(),
                            port,
                            username,
                            userSshKeyInfo,
                            proxyInfo.getHostKey(),
                            true /*limitName*/);

                    //TODO handle if 1st port is not reachable, but others are (login and GetVersion test should pick another port in that case)
                    //test login on first port
                    SSHLoginTest proxyLoginTest = new SSHLoginTest(
                            proxyInfo.getHostname(), port,
                            conProxyInfo.getUsername() /*don't use username directly: SshProxyInfo constructor shortens it if needed*/,
                            null,/*no proxy: this is a login to the proxy itself*/geniUserProvider,
                            jFedPreferences);
                    tests.add(proxyLoginTest);
//
                    URL amUrl = auth.getUrl(ServerType.GeniServerRole.AM, 3);
                    int version = 3;
                    if (amUrl == null && auth.getUrl(ServerType.GeniServerRole.AM, 2) != null) {
                        amUrl = auth.getUrl(ServerType.GeniServerRole.AM, 2);
                        version = 2;
                    }
                    //fall back to to wall2 AMv3
                    if (amUrl == null) {
                        auth =
                                authorityListModel.getByUrnExact(WALL2_URN);
                        if (auth != null) {
                            amUrl = auth.getUrl(ServerType.GeniServerRole.AM, 3);
                            version = 3;
                        }
                    }
                    if (amUrl != null) {
                        GetVersionTest getVersionTest =
                                connectivityTestFactory.createGetVersionTest(auth, conProxyInfo, version);
                        tests.add(getVersionTest);
                    }
                }

                //test maximum 10 ports for the same proxy
                int maxTests = 10;
                for (Integer curPort : portRange) {
                    if (maxTests-- <= 0)
                        break;


                    final HostAndPortTest proxyTest = new HostAndPortTest(proxyInfo.getHostname(), curPort,
                            "Proxy " + proxyInfo.getHostname() + ":" + curPort, "Proxy"/*category*/);

                    if (!tests.contains(proxyTest))
                        tests.add(proxyTest);
                }

            }
        }
        return tests;
    }

}
