package be.iminds.ilabt.jfed.experimenter_gui.canvas;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.CanvasNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasNode;
import be.iminds.ilabt.jfed.experimenter_gui.editor.NewCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.views.DraggedItemProvider;
import be.iminds.ilabt.jfed.experimenter_gui.util.ExecuteOnNotNull;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.ReadOnlyListWrapper;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.Scene;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;
import javafx.scene.input.MouseDragEvent;
import javafx.scene.input.MouseEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * User: twalcari
 * Date: 5/21/2015
 * Time: 10:27
 */
public abstract class AbstractEditableCanvas extends ExperimentCanvas {
    private static final Logger LOG = LoggerFactory.getLogger(AbstractEditableCanvas.class);

    protected static final double EDITOR_LOCATION_SHIFT = 15;

    protected final ObservableList<CanvasNode> clipboard = FXCollections.observableArrayList();
    private final StringProperty status = new SimpleStringProperty();

    private final NewCanvasLink newCanvasLink;
    private final DraggedItemProvider draggedItemProvider;

    public AbstractEditableCanvas(FXModelRspec model, JFedConfiguration config, CanvasNodeFactory canvasNodeFactory,
                                  DraggedItemProvider draggedItemProvider, AuthorityList authorityList) {
        super(model, config, canvasNodeFactory);
        this.draggedItemProvider = draggedItemProvider;

        this.newCanvasLink = new NewCanvasLink(this, authorityList);

        this.setOnMouseDragReleased(this::onMouseDragReleased);

        //make sure that the initialize-function is called when the scene-property is set
        new ExecuteOnNotNull<Scene>(sceneProperty()) {
            @Override
            public void run() {
                initialize();
            }
        };
    }

    private void initialize() {
        //add mnemorics
        getScene().getAccelerators().put(new KeyCodeCombination(KeyCode.DELETE), this::deleteSelectedItem);
    }

    protected abstract void deleteSelectedItem();


    @Override
    protected void canvasNodeOnDragDetected(MouseEvent mouseEvent) {
        super.canvasNodeOnDragDetected(mouseEvent);

        //make canvas link invisible when starting to drag
        newCanvasLink.setVisible(false);
    }

    @Override
    protected void canvasNodeOnMouseDragged(MouseEvent mouseEvent) {
        CanvasNode node = (CanvasNode) mouseEvent.getSource();
        Bounds canvasOrigin = getCanvas().
                localToScene(getCanvas().getBoundsInLocal());
        double newX = mouseEvent.getSceneX()
                - canvasOrigin.getMinX()
                - node.getBoundsInParent().getWidth() / 2;
        double newY = mouseEvent.getSceneY()
                - canvasOrigin.getMinY()
                - node.getBoundsInParent().getHeight() / 2;

        //setSecondaryStatus("Moving to " + newX + ", " + newY);

        node.relocate(newX, newY);

        mouseEvent.consume();
    }

    /**
     * Performs panning of the canvas when the newCanvasLink is not visible, or when the shift-button is down.
     * Otherwise, it tries to connect two nodes
     */
    protected void onCanvasMouseDragged(MouseEvent mouseEvent) {
        if (mouseEvent.isShiftDown() || !newCanvasLink.isVisible()) {
            super.onCanvasMouseDragged(mouseEvent);
        } else {

            //try to connect this node if canvas isn't fixed
            RspecCanvasNode targetNode = findNearbyNode(mouseEvent.getX(), mouseEvent.getY());

            if (targetNode != null && targetNode != newCanvasLink.getOriginNode()) {
                newCanvasLink.setTargetNode(targetNode);
            } else {

                RspecCanvasLink targetLink = findNearbyLink(mouseEvent.getX(), mouseEvent.getY());

                if (targetLink != null) {
                    newCanvasLink.setTargetLink(targetLink);
                } else {
                    newCanvasLink.setNewEnd(mouseEvent.getX(), mouseEvent.getY());
                }
            }
        }
    }

    @Override
    protected void onCanvasMouseReleased(MouseEvent mouseEvent) {
        super.onCanvasMouseReleased(mouseEvent);

        //have we made a new connection?
        if (newCanvasLink.isLinkComplete()) {
            //check if new link may be created

            if (!newCanvasLink.isLegalLink()) {
                setStatus("The requested link is illegal and was ignored.");
                LOG.trace("Link is not legal, and is thus rejected.");
            } else {

                processNewLinkRequest(newCanvasLink);
            }
        }
        newCanvasLink.clear();
        newCanvasLink.setVisible(false);
    }

    @Override
    protected void onCanvasMouseMoved(MouseEvent mouseEvent) {
        if (!mouseEvent.isPrimaryButtonDown()) {

            //if the mouse is close nearby a node, show a "newCanvasLink"

            RspecCanvasNode nearbyNode = findNearbyNode(mouseEvent.getX(), mouseEvent.getY());
            RspecCanvasLink nearbyLink = null;
            if (nearbyNode != null) {
                newCanvasLink.setOriginNode(nearbyNode);
            } else {
                nearbyLink = findNearbyLink(mouseEvent.getX(), mouseEvent.getY());
                if (nearbyLink != null) {
                    newCanvasLink.setOriginLink(nearbyLink);
                }
            }

            if (nearbyNode != null || nearbyLink != null) {
                newCanvasLink.setNewEnd(mouseEvent.getX(), mouseEvent.getY());
                newCanvasLink.setVisible(true);

                setStatus("Click and drag the circle to a target node to create a new link.");
            } else {

                newCanvasLink.setVisible(false);
            }
        }
    }

    protected abstract boolean processNewLinkRequest(NewCanvasLink newCanvasLink);


    public void onMouseDragReleased(MouseDragEvent mouseDragEvent) {
        //we are adding a new element from the toolbox
        assert (draggedItemProvider.getDraggedItem() != null);

        Point2D location = new Point2D(
                mouseDragEvent.getSceneX()
                        - getCanvas().getLocalToSceneTransform().getTx()
                        - draggedItemProvider.getDraggedItem().getPrefWidth() / 2,
                mouseDragEvent.getSceneY()
                        - getCanvas().getLocalToSceneTransform().getTy()
                        - draggedItemProvider.getDraggedItem().getPrefHeight() / 2
        );

        addCanvasNodeFromToolboxItem(draggedItemProvider.getDraggedItem(), location);

        draggedItemProvider.clearDraggedItem();
        setCursor(null);
        mouseDragEvent.consume();
    }

    protected abstract void addCanvasNodeFromToolboxItem(ToolboxItem item, Point2D location);


    public void copySelected() {
        clipboard.clear();

        if (getSelectionProvider().getSelectedCanvasNode() == null) {
            JFDialogs.create()
                    .owner(this.getScene().getWindow())
                    .message("Nothing to copy!\r\nPlease select an item first.")
                    .masthead("Copy action failed")
                    .showWarning();
        }

        clipboard.add(getSelectionProvider().getSelectedCanvasNode());
    }

    public final void paste() {
        if (clipboard.isEmpty()) {
            JFDialogs.create()
                    .owner(this.getScene().getWindow())
                    .message("Nothing to paste!\r\nPlease copy an item first.")
                    .masthead("Paste action failed")
                    .showWarning();
        }

        /* If the user pastes the same node two times, we want it to appear on two different places.
           Therefore we modify the clipboard to the newly created items    */
        List<CanvasNode> newClipboard = new ArrayList<>();

        //we now know that a RSpecCanvasNode has been created, and we can thus safely create the interfaces
        clipboard.stream()
                .filter(canvasNode -> canvasNode instanceof RspecCanvasNode)
                .forEach(canvasNode -> {
                    RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) canvasNode;
                    RspecCanvasNode newRspecNode = addCopyOfNodeToCanvas(rspecCanvasNode);

                    newClipboard.add(newRspecNode);
                });

        clipboard.clear();
        clipboard.addAll(newClipboard);


    }

    public void duplicateSelected() {
        if (getSelectionProvider().getSelectedCanvasNode() != null &&
                getSelectionProvider().getSelectedCanvasNode() instanceof RspecCanvasNode) {
            RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) getSelectionProvider().getSelectedCanvasNode();

            RspecCanvasNode newRspecCanvasNode = addCopyOfNodeToCanvas(rspecCanvasNode);
            getSelectionProvider().setSelectedCanvasNode(newRspecCanvasNode);
        } else {
            JFDialogs.create()
                    .owner(this.getScene().getWindow())
                    .message("Nothing to duplicate!\r\nPlease select an item first.")
                    .masthead("Duplicate action failed")
                    .showWarning();

        }
    }

    protected abstract RspecCanvasNode addCopyOfNodeToCanvas(RspecCanvasNode node);

    public ReadOnlyListProperty<CanvasNode> getClipboard() {
        return new ReadOnlyListWrapper<>(clipboard).getReadOnlyProperty();
    }

    public String getStatus() {
        return status.get();
    }

    public StringProperty statusProperty() {
        return status;
    }

    protected void setStatus(String status) {
        if (Objects.equals(this.status.get(), status))
            return; //ignore an unchanged status message

        LOG.trace("New Status: {}", status);
        this.status.set(status);
    }
}
