package be.iminds.ilabt.jfed.experimenter_gui.controller.model;

import be.iminds.ilabt.jfed.experimenter_gui.controller.view.ControllerNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.util.ExecuteOnNotNull;
import be.iminds.ilabt.jfed.experimenter_gui.util.ProxyServiceUtil;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.controller.CommandScheduler;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableBooleanValue;
import javafx.beans.value.ObservableValue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/*
 * The TimelineModel will keep track of timestamps, nodes and their
 * added tests. 
 * 
 * The TimelineModel will keep a thread running for updating its time. 
 * The view will asynchronously update its timeline on a different interval
 */
public class ControllerModel {
    private static final Logger LOG = LoggerFactory.getLogger(ControllerModel.class);

    private final GeniUserProvider geniUserProvider;

    private final SshConnectionPool sshConnectionPool;
    private final BooleanProperty stopped = new SimpleBooleanProperty(true);
    private final BooleanProperty started = new SimpleBooleanProperty(false);
    private final BooleanProperty repeating = new SimpleBooleanProperty(false);
    //Map barrierSegmentID to schedulers
    //A TreeMap will sort according to the natural ordering of its keys
    private final Map<Integer, CommandScheduler> schedulers = new TreeMap<>();


    // This FXModelRspec will get node updates: nodes might be added, and login info may be added/removed
    private final ObjectProperty<FXModelRspec> nodeInfoModelRspecProperty = new SimpleObjectProperty<>();

    // This FXModelRspec is to be used to store the updated commands and barriers
    private final ObjectProperty<FXModelRspec> commandsModelRspecProperty = new SimpleObjectProperty<>();

    //assumptions: if nodes are added and/or removed, these 2 models are the same object!
    //             if these 2 are seperate, then only the node login info (and other info) is updated in nodeInfoModelRspecProperty
    //                and the basic node count and info stays the same in both
    //             both models can be changed by external parties (following the rules above)

    private final ReadOnlyBooleanWrapper sshConnectionInitStarted = new ReadOnlyBooleanWrapper(false);
    /**
     * TRUE if this controller is used in the editor-phase, FALSE if it is used in the viewer-phase
     */
    private final ReadOnlyObjectWrapper<CommandEditorMode> editorMode =
            new ReadOnlyObjectWrapper<>(CommandEditorMode.OFFLINE_EDITOR);
    private final Map<ExperimentCommand, Thread> instantCommands = new HashMap<>();
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final ProxyServiceUtil proxyServiceUtil;
    private final ControllerNodeFactory controllerNodeFactory;

    //The slice and experimenterModel belonging to this experiment

    private Experiment experiment;

    private int currentCommandID = 0;
    private final ChangeListener<Boolean> barrierSegmentCompletedListener = new ChangeListener<Boolean>() {

        @Override
        public void changed(ObservableValue<? extends Boolean> ov, Boolean t, Boolean t1) {
            if (t1) {
                currentBarrierSegment.completedProperty().removeListener(this);
                schedulers.get(currentBarrierSegment.getOrderNumber()).stopTime();
                ExperimentBarrierSegment nextBarrierSegment = getNextBarrierSegment();
                //Stop the experiment if all barrier segments have been completed
                //If there are segments left, select the next, add the listener and start it
                if (nextBarrierSegment == null) {
                    stopTime();

                    if (repeating.get()) {
                        rewindTime();
                        startTime();
                    }
                } else {
                    elapsedTime.set(0);
                    startTime = System.currentTimeMillis();
                    currentBarrierSegment = nextBarrierSegment;
                    currentBarrierSegment.completedProperty().addListener(barrierSegmentCompletedListener);
                    currentBarrierSegment.start();
                    if (!stopped.get())
                        schedulers.get(currentBarrierSegment.getOrderNumber()).startTime();
                }
            }
        }

    };
    private int nextBarrierID = 0;
    private long startTime = 0;
    private LongProperty elapsedTime = new SimpleLongProperty(0);
    private ExperimentBarrierSegment currentBarrierSegment;
    private int instantCommandCount = 0;
    private int commandCount = 0;

    //instance initializer
    {
        started.bind(stopped.not());
    }

//    private SliceController sliceController = null;

    /**
     * @param editorMode              Defines if this model is used for editorMode editing or online editing/running of the experiment
     * @param geniUserProvider
     * @param proxyPreferencesManager
     * @param proxyServiceUtil
     * @param controllerNodeFactory
     */
    public ControllerModel(FXModelRspec requestModelRspec, CommandEditorMode editorMode,
                           GeniUserProvider geniUserProvider, ProxyPreferencesManager proxyPreferencesManager,
                           ProxyServiceUtil proxyServiceUtil, ControllerNodeFactory controllerNodeFactory) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.proxyServiceUtil = proxyServiceUtil;
        this.controllerNodeFactory = controllerNodeFactory;
        assert requestModelRspec != null;
        //should be this.commandsModelRspecProperty.setValue(FXRspecFactory.getInstance().copyModelRspec(requestModelRspec));
        this.commandsModelRspecProperty.setValue(requestModelRspec);
        this.nodeInfoModelRspecProperty.setValue(requestModelRspec); //not yet known. goOnline will set it.
        this.editorMode.set(editorMode);

        this.sshConnectionPool = new SshConnectionPool(this.nodeInfoModelRspecProperty,
                geniUserProvider.getLoggedInGeniUser(), this.proxyPreferencesManager, this.proxyServiceUtil);

        //init any commands that were already present in the modelRspec
        initExistingCommands();
    }

    /**
     * @param modelRspecEditor
     * @param editorMode              Defines if this model is used for editorMode editing or online editing/running of the experiment
     * @param geniUserProvider
     * @param proxyPreferencesManager
     * @param proxyServiceUtil
     * @param controllerNodeFactory
     */
    public ControllerModel(ModelRspecEditor modelRspecEditor, CommandEditorMode editorMode,
                           GeniUserProvider geniUserProvider, ProxyPreferencesManager proxyPreferencesManager,
                           ProxyServiceUtil proxyServiceUtil, ControllerNodeFactory controllerNodeFactory) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.proxyServiceUtil = proxyServiceUtil;
        this.controllerNodeFactory = controllerNodeFactory;
        //both models are the same
        this.nodeInfoModelRspecProperty.setValue(modelRspecEditor.getModelRspec());
        this.commandsModelRspecProperty.bind(nodeInfoModelRspecProperty);

        this.editorMode.set(editorMode);

        this.sshConnectionPool = new SshConnectionPool(nodeInfoModelRspecProperty,
                geniUserProvider.getLoggedInGeniUser(), this.proxyPreferencesManager, this.proxyServiceUtil);

        //init any commands that were already present in the modelRspec
        initExistingCommands();
    }

    private final ObjectProperty<ManifestRspecSource> manifestRspecSourceProperty = new SimpleObjectProperty<>();

    //initialize fixed listener
    {
        manifestRspecSourceProperty.addListener(new ChangeListener<ManifestRspecSource>() {
            @Override
            public void changed(ObservableValue<? extends ManifestRspecSource> observableValue,
                                ManifestRspecSource oldManifestRspecSource,
                                ManifestRspecSource newManifestRspecSource) {
                ManifestRspecSource source = newManifestRspecSource;
                if (source == null) {
                    ControllerModel.this.nodeInfoModelRspecProperty.setValue(null);
                } else {
                    ControllerModel.this.nodeInfoModelRspecProperty.setValue((FXModelRspec) source.getModelRspec(ModelRspecType.FX));
                }
            }
        });
    }

    /**
     * Connects this model to a slice
     * <p>
     * Makes everything ready to create SSH connections. But does not yet create them.
     *
     * @param experiment
     */
    public void setExperiment(final Experiment experiment) {
        this.experiment = experiment;

        //todo: bind to status instead
        experiment.deletingProperty().addListener(new ChangeListener<Boolean>() {
            @Override
            public void changed(ObservableValue<? extends Boolean> observableValue, Boolean oldBoolean, Boolean newBoolean) {
                if (newBoolean)
                    goOffline();
            }
        });

        new ExecuteOnNotNull<Slice>(experiment.sliceProperty()) {
            @Override
            public void run() {
                manifestRspecSourceProperty.bind(experiment.getSlice().manifestRspecProperty());
            }
        };
    }

    private void goOffline() {
        this.experiment = null;
        manifestRspecSourceProperty.unbind();
        manifestRspecSourceProperty.setValue(null);
    }

    public SshConnectionPool getSshConnectionPool() {
        return sshConnectionPool;
    }

    public ObservableBooleanValue allConnectableProperty() {
        return sshConnectionPool.allConnectableProperty();
    }

    public ObservableBooleanValue allConnectedProperty() {
        return sshConnectionPool.allConnectedProperty();
    }

    public ObservableBooleanValue anyConnectableProperty() {
        return sshConnectionPool.anyConnectableProperty();
    }

    public ObservableBooleanValue anyConnectedProperty() {
        return sshConnectionPool.anyConnectedProperty();
    }

    /**
     * Retrieves the commands and barriersegments from the model
     * (which might have been created from an RSpec) and adds them to this model
     */
    private void initExistingCommands() {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        if (modelRspec == null) return;

        for (ExperimentBarrierSegment b : modelRspec.getBarrierSegments()) {
            schedulers.put(b.getOrderNumber(), new CommandScheduler(b));
            if (b.getOrderNumber() >= nextBarrierID)
                nextBarrierID = b.getOrderNumber() + 1;
        }

        if (!modelRspec.getBarrierSegments().isEmpty())
            currentBarrierSegment = modelRspec.getBarrierSegments().get(0);

        for (FXRspecNode n : modelRspec.getNodes()) {
            for (ExperimentCommand comm : n.getCommands()) {
//                if (!schedulers.containsKey(comm.getBarrierSegmentOrderNumber())) {
//                    LOG.error("ExperimentCommand references BarrierSegmentOrderNumber="+comm.getBarrierSegmentOrderNumber()+
//                            " but that BarrierSegment was not declared. Command will be ignored!");
//                    continue;
//                }
                comm.setID(currentCommandID++);


                //// handle missing barriersegments by creating them ////
                int barrierNr = comm.getBarrierSegmentOrderNumber();
                CommandScheduler commandScheduler = schedulers.get(barrierNr);
                if (commandScheduler == null) {
                    LOG.warn("Did not find CommandScheduler with orderNr " + comm.getBarrierSegmentOrderNumber() + " will create one.");
                    ExperimentBarrierSegment barrierSegment =
                            new ExperimentBarrierSegment("Barrier segment " + barrierNr, barrierNr);
                    assert barrierSegment.getOrderNumber() == barrierNr;
                    modelRspec.addBarrierSegment(barrierSegment);

                    commandScheduler = new CommandScheduler(barrierSegment);
                    schedulers.put(barrierNr, commandScheduler);

                    if (currentBarrierSegment == null)
                        currentBarrierSegment = barrierSegment;
                } else {
                    //LOG.debug("Found CommandScheduler");
                }
                ////

                commandScheduler.addCommand(comm);
                comm.setSshConnectionProvider(sshConnectionPool);
                commandCount++;
            }
        }
    }

    /**
     * @return List<FXRspecNode> from modelRspec
     */
    public List<FXRspecNode> getNodes() {
        if (nodeInfoModelRspecProperty.get() == null) return Collections.emptyList();
        return nodeInfoModelRspecProperty.get().getNodes();
    }

    /**
     * Starts the experiment. It will continue from the current elapsedTime.
     */
    public void startTime() {
        stopped.set(false);
        startTime = System.currentTimeMillis() - elapsedTime.get();

        if (currentBarrierSegment == null)
            currentBarrierSegment = getNextBarrierSegment();

        if (currentBarrierSegment != null) {
            currentBarrierSegment.completedProperty().addListener(barrierSegmentCompletedListener);
            currentBarrierSegment.start();
            schedulers.get(currentBarrierSegment.getOrderNumber()).startTime();
        }
    }

    /**
     * Sets the elapsed time at the requested barrierSegment
     *
     * @param t new elapsed time
     * @param b barrier segment to set the time at
     */
    public void setElapsedTime(long t, ExperimentBarrierSegment b) {
        //Check if the time has to be restarted after setting it
        boolean restart = !stopped.get();

        //Rewind, this will stop the time and rewind to the first barrier segment, with all commands/barriers put in the unstarted state
        rewindTime();

        //Skip to the right barrier segment
        while (currentBarrierSegment != b) {
            if (currentBarrierSegment != null)
                currentBarrierSegment.skip();
            currentBarrierSegment = getNextBarrierSegment();
        }

        //Set the time
        elapsedTime.set(t);
        startTime = System.currentTimeMillis() - elapsedTime.get();

        currentBarrierSegment.completedProperty().addListener(barrierSegmentCompletedListener);
        currentBarrierSegment.start();

        schedulers.get(currentBarrierSegment.getOrderNumber()).setElapsedTime(t);

        if (restart) {
            //Start the time again
            stopped.set(false);
            schedulers.get(currentBarrierSegment.getOrderNumber()).startTime();
        }
    }

    /**
     * Stops the experiment run. It can be continued by using startTime.
     */
    public void stopTime() {
        schedulers.get(currentBarrierSegment.getOrderNumber()).stopTime();
        currentBarrierSegment.completedProperty().removeListener(barrierSegmentCompletedListener);
        stopped.set(true);
    }

    /**
     * Rewinds the experiment to the first barrier segment and to elapsed time 0.
     * If no barrier segments exist in the experiment, then the current barrier
     * segment will be null.
     */
    public void rewindTime() {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        for (CommandScheduler scheduler : schedulers.values()) {
            scheduler.rewind();
        }

        for (ExperimentBarrierSegment b : modelRspec.getBarrierSegments()) {
            b.rewind();
        }

        startTime = 0;
        elapsedTime.set(0);
        stopped.set(true);

        if (currentBarrierSegment != null) {
            currentBarrierSegment.completedProperty().removeListener(barrierSegmentCompletedListener);
        }

        if (modelRspec.getBarrierSegments().isEmpty()) {
            currentBarrierSegment = null;
        } else {
            currentBarrierSegment = modelRspec.getBarrierSegments().get(0);
        }
    }

    /**
     * @return current elapsed time
     */
    public long getElapsedTime() {
        if (!stopped.get()) {
            elapsedTime.set(System.currentTimeMillis() - startTime);
        }
        return elapsedTime.get();
    }

    /**
     * @return a BooleanProperty representing the paused state of the experiment
     */
    public BooleanProperty getStoppedProperty() {
        return stopped;
    }

    /**
     * Adds the given command to the experiment
     *
     * @param comm
     */
    public void addCommand(ExperimentCommand comm) {
        comm.setSshConnectionProvider(sshConnectionPool);
        comm.setID(currentCommandID++);

        schedulers.get(comm.getBarrierSegmentOrderNumber()).addCommand(comm);

        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        FXRspecNode n = modelRspec.getNodeByUniqueId(comm.getNodeUniqueID());
        if (n != null) {
            n.getCommands().add(comm);
        } else {
            LOG.error("Failed to add command " + comm.getTag() + "to node " + comm.getNodeUniqueID() + ": node does not exist in modelRspec.");
        }

        ExperimentBarrierSegment b = modelRspec.getBarrierSegmentById(comm.getBarrierSegmentOrderNumber());
        if (b != null) {
            if (b.getOrderNumber() < currentBarrierSegment.getOrderNumber()) {
                comm.setSkipped(true);
                comm.setStarted(true);
                comm.setFinished(true);
            }

            b.addCommand(comm);
        } else {
            LOG.error("Failed to add command " + comm.getTag() + "to barrier segment " + comm.getBarrierSegmentOrderNumber() + ": barrier segment does not exist in modelRspec.");
        }
    }

    /**
     * A barrier segment gets added to all the nodes, with 1 unique barrierSegmentID. This ID is set by the ControllerModel.
     * The nodeID is set by the model as well, while it adds the barrier segments to the nodes.
     *
     * @param b
     */
    public void addBarrierSegment(ExperimentBarrierSegment b) {
        b.setOrderNumber(nextBarrierID++);
        if (b.getTag().equals(""))
            b.setTag("Barrier segment " + b.getOrderNumber());

        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        modelRspec.addBarrierSegment(b);

        schedulers.put(b.getOrderNumber(), new CommandScheduler(b));

        if (currentBarrierSegment == null)
            currentBarrierSegment = modelRspec.getBarrierSegments().get(0);
    }

    /**
     * Deletes the requested barrier segment from the experiment. All the commands
     * belonging to this segment will be deleted as well.
     *
     * @param b barrier segment
     */
    public void deleteBarrierSegment(ExperimentBarrierSegment b) {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        if (modelRspec.getBarrierSegments().contains(b)) {
            //If the current gets deleted, handle selecting + starting the new segment
            if (currentBarrierSegment == b) {
                currentBarrierSegment.completedProperty().removeListener(barrierSegmentCompletedListener);
                currentBarrierSegment = getNextBarrierSegment();

                if (!stopped.get()) {
                    if (currentBarrierSegment != null) {
                        currentBarrierSegment.completedProperty().addListener(barrierSegmentCompletedListener);
                        currentBarrierSegment.start();
                        schedulers.get(currentBarrierSegment.getOrderNumber()).startTime();
                    } else {
                        stopped.set(true);
                    }
                }
            }

            //Remove the correct segment
            modelRspec.getBarrierSegments().remove(b);
            schedulers.get(b.getOrderNumber()).stopTime();
            schedulers.get(b.getOrderNumber()).unregister();
            schedulers.remove(b.getOrderNumber());
        }
    }

    /**
     * @return The next barrier segment, or null if no next barrier segment exists
     */
    public ExperimentBarrierSegment getNextBarrierSegment() {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        if (currentBarrierSegment == null) {
            if (modelRspec.getBarrierSegments().isEmpty())
                return null;
            else return modelRspec.getBarrierSegments().get(0);
        } else {
            int currentIndex = modelRspec.getBarrierSegments().lastIndexOf(currentBarrierSegment);
            if (currentIndex + 1 < modelRspec.getBarrierSegments().size()) {
                return modelRspec.getBarrierSegments().get(currentIndex + 1);
            } else {
                return null;
            }
        }
    }

    /**
     * @return The previous barrier segment, or null if no previous barrier segment exists
     */
    public ExperimentBarrierSegment getPreviousBarrierSegment() {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        if (currentBarrierSegment == null) {
            if (modelRspec.getBarrierSegments().isEmpty())
                return null;
            else return modelRspec.getBarrierSegments().get(modelRspec.getBarrierSegments().size() - 1);
        } else {
            int currentIndex = modelRspec.getBarrierSegments().lastIndexOf(currentBarrierSegment);
            if (currentIndex - 1 > 0) {
                return modelRspec.getBarrierSegments().get(currentIndex - 1);
            } else {
                return null;
            }
        }
    }

    /**
     * @return commandCount +1, used for automatically assigning tags to a command
     */
    public int nextCommandNumber() {
        return ++commandCount;
    }

    /**
     * Execute a command as an instant command
     *
     * @param comm
     */
    public void instantCommand(ExperimentCommand comm) {
        comm.setSshConnectionProvider(sshConnectionPool);
        Thread t = new Thread(comm);
        t.start();

        instantCommands.put(comm, t);
    }

    /**
     * Stop an instant command
     *
     * @param command
     */
    public void stopInstantCommand(ExperimentCommand command) {
        instantCommands.get(command).interrupt();
    }

    public int nextInstantCommandNumber() {
        return ++instantCommandCount;
    }

    /**
     * Deletes the requested command from the experiment
     *
     * @param comm
     */
    public void deleteCommand(ExperimentCommand comm) {
        schedulers.get(comm.getBarrierSegmentOrderNumber()).deleteCommand(comm);

        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        modelRspec.getNodeByUniqueId(comm.getNodeUniqueID()).getCommands().remove(comm);
        modelRspec.getBarrierSegmentById(comm.getBarrierSegmentOrderNumber()).deleteCommand(comm);
    }

    public ExperimentCommand getCommand(int commandID) {
        FXModelRspec modelRspec = commandsModelRspecProperty.get();
        for (FXRspecNode node : modelRspec.getNodes())
            for (ExperimentCommand comm : node.getCommands()) {
                if (comm.getID() == commandID) {
                    return comm;
                }
            }
        return null;
    }

    public FXRspecNode getRspecNode(String uniqueNodeID) {
        for (FXRspecNode n : getNodes()) {
            if (n.getUniqueId().equals(uniqueNodeID)) {
                return n;
            }
        }
        return null;
    }

    public FXModelRspec getCommandModelRspec() {
        return commandsModelRspecProperty.get();
    }

    /**
     * True: this model is for editing the experiment only, not for running it
     * False: this model is to be used with real nodes
     *
     * @return
     */
    public ReadOnlyObjectProperty<CommandEditorMode> editorModeProperty() {
        return editorMode.getReadOnlyProperty();
    }

    /**
     * Calls unregister on all the command schedulers.
     * This will make sure no changeListeners remain bound to commands.
     */
    public void unregister() {
        for (CommandScheduler cs : schedulers.values()) {
            cs.unregister();
        }
    }

    public BooleanProperty stoppedProperty() {
        return stopped;
    }

    public boolean isStopped() {
        return stopped.get();
    }

    /**
     * @return a boolean representing the running state of the experiment
     */
    public boolean isStarted() {
        return started.get();
    }

    public BooleanProperty startedProperty() {
        return started;
    }

    public boolean getRepeating() {
        return repeating.get();
    }

    public BooleanProperty repeatingProperty() {
        return repeating;
    }

    public void setRepeating(boolean repeating) {
        this.repeating.set(repeating);
    }

    public LongProperty elapsedTimeProperty() {
        return elapsedTime;
    }

    public CommandEditorMode getEditorMode() {
        return editorMode.get();
    }

    public ControllerNodeFactory getControllerNodeFactory() {
        return controllerNodeFactory;
    }

    public Experiment getExperiment() {
        return experiment;
    }


    public static enum CommandEditorMode {OFFLINE_EDITOR, ONLINE_EDITOR}
}
