package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.WaitForTaskExecutionsTask;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSliceMembersTask;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSubAuthorityMembersTask;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxListCell;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import javafx.util.Callback;
import javafx.util.StringConverter;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 */
public class ShareSliceDialog extends VBox implements Consumer<ActionEvent> {
    private static final Logger LOG = LoggerFactory.getLogger(ShareSliceDialog.class);
    private static final StringConverter<GeniUrn> URN_TO_RESOURCENAME_STRINGCONVERTER = new StringConverter<GeniUrn>() {
        @Override
        public String toString(GeniUrn urn) {
            return urn.getResourceName();
        }

        @Override
        public GeniUrn fromString(String string) {
            throw new UnsupportedOperationException();
        }
    };

    private final Slice slice;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;

    private final GeniUserProvider geniUserProvider;
    private final EasyModel easyModel;
    private final HighLevelController hlc;


    @FXML
    private ListView<SelectableUser> usersList;

    @FXML
    private HBox initializingBox;

    @FXML
    private CheckBox addSshKeysCheckBox;

    @FXML
    private Label helpLabel;

    private final org.controlsfx.dialog.Dialog dialog;

    ShareSliceDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory,
                     GeniUserProvider geniUserProvider, EasyModel easyModel, HighLevelController hlc) {
        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.geniUserProvider = geniUserProvider;
        this.easyModel = easyModel;
        this.hlc = hlc;
        this.slice = experimentController.getExperiment().getSlice();

        //initialize the Scene
        FXMLUtil.injectFXML(this);


        usersList.setCellFactory(new Callback<ListView<SelectableUser>, ListCell<SelectableUser>>() {
            @Override
            public ListCell<SelectableUser> call(ListView<SelectableUser> param) {
                return new CheckBoxListCell<SelectableUser>(SelectableUser::selectedProperty) {
                    @Override
                    public void updateItem(SelectableUser item, boolean empty) {
                        super.updateItem(item, empty);

                        if (item != null) {
                            setDisable(item.isFixed());
                        }
                    }
                };
            }
        });

        WaitForTaskExecutionsTask waitTask = new WaitForTaskExecutionsTask() {

            @Override
            public void onTaskExecutionsFinished() {
                updateUsersList();

                initializingBox.setVisible(false);
            }
        };

        if (experimentController.getExperiment().getSlice().getUsers().isEmpty()) {
            ListSliceMembersTask listSliceMembersTask =
                    hlc.listSliceMembers(experimentController.getExperiment().getSlice());

            waitTask.addTaskExecution(hlc.submitTask(listSliceMembersTask));
        }

        if (easyModel.getUsersBySubAuthority(experimentController.getExperiment().getSubAuthName()).isEmpty()) {
            ListSubAuthorityMembersTask listSubAuthorityMembersTask =
                    hlc.listSubAuthorityMembers(experimentController.getExperiment().getSubAuthName());

            waitTask.addTaskExecution(hlc.submitTask(listSubAuthorityMembersTask));
        }


        new Thread(null, waitTask, "ShareSlice-WaitForBTaskExecutionsTasks").start();

        final String sliceProject = slice.getSubAuthorityName();
        final String authName = geniUserProvider.getLoggedInGeniUser().getUserAuthority().getName();

        String helpText =
                "Sharing an experiment has some limitations:\n" +
                        "- You can only share with users of the same authority (" + authName + ")\n";

        if (sliceProject != null)
            helpText +=
                    "- You can only share with users that are member of the \"" + sliceProject + "\" project in which this experiment was created.\n";

        helpLabel.setText(helpText);

        //initialize dialog
        dialog = new org.controlsfx.dialog.Dialog(parentWindow, "Share Experiment");
        dialog.setMasthead("Share your experiment");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialog.setContent(this);
        dialog.getActions().setAll(new DialogAction("Share", org.controlsfx.control.ButtonBar.ButtonType.OK_DONE, false, false, true, this),
                org.controlsfx.dialog.Dialog.ACTION_CANCEL);

    }

    @FXML
    private void addCustomUser() {

        TextInputDialog dialog = new TextInputDialog();

        dialog.setHeaderText("Share experiment to custom user");
        dialog.setContentText("Username:");

        dialog.getEditor().textProperty().addListener((observable, oldValue, newValue) -> {
            String strippedValue = newValue.replaceAll("[^A-Za-z0-9_-]", "");

            if (!strippedValue.equals(newValue)) {
                dialog.getEditor().setText(strippedValue);
            }
        });


        dialog.showAndWait().ifPresent(username -> {
            final GeniUrn curUserUrn = geniUserProvider.getLoggedInGeniUser().getUserUrn();
            GeniUrn newUserUrn =
                    GeniUrn.createGeniUrnFromEncodedParts(curUserUrn.getEncodedTopLevelAuthority(), "user", username);

            Optional<SelectableUser> existingUser =
                    usersList.getItems().stream().filter(su -> su.getUrn().equals(newUserUrn)).findFirst();

            if (!existingUser.isPresent()) {
                final SelectableUser newSelectableUser = new SelectableUser(newUserUrn, true, false);
                usersList.getItems().add(newSelectableUser);
                usersList.scrollTo(newSelectableUser);
            } else {
                new Alert(Alert.AlertType.ERROR, "The user already exists in the list!").showAndWait();
                usersList.scrollTo(existingUser.get());
            }

        });

    }

    private void updateUsersList() {
        assert !experimentController.getExperiment().getSlice().getUsers().isEmpty();
        assert !easyModel.getUsersBySubAuthority(experimentController.getExperiment().getSubAuthName()).isEmpty();

        usersList.getItems().clear();

        final Collection<GeniUrn> currentMembers = experimentController.getExperiment().getSlice().getUsers();
        for (GeniUrn currentMember : currentMembers) {
            usersList.getItems().addAll(new SelectableUser(currentMember, true, true));
        }

        List<GeniUrn> candidateMembers =
                new ArrayList<>(easyModel.getUsersBySubAuthority(experimentController.getExperiment().getSubAuthName()));
        candidateMembers.removeAll(currentMembers);

        for (GeniUrn candidateMember : candidateMembers) {
            usersList.getItems().addAll(new SelectableUser(candidateMember, false, false));
        }

        usersList.getItems().sort(Comparator.comparing(SelectableUser::isFixed).reversed()
                .thenComparing(SelectableUser::getSelected)
                .thenComparing(su -> su.getUrn().getResourceName().toLowerCase()));

    }


    public void showDialog() {
        dialog.show();
    }

    @Override
    public void accept(ActionEvent actionEvent) {
        List<GeniUrn> selectedUsers = usersList.getItems().stream()
                .filter(SelectableUser::getSelected)
                .map(SelectableUser::getUrn)
                .collect(Collectors.toList());

        //remove all existing users
        selectedUsers.removeAll(experimentController.getExperiment().getSlice().getUsers());

        if (selectedUsers.isEmpty()) {
            new Alert(Alert.AlertType.ERROR, "You need to select at least one new project member.", ButtonType.OK).showAndWait();

            dialog.shake();
        } else {
            experimentController.submitExperimentTaskGenerator(
                    experimentTasksFactory.createShareSliceTasksFactory(experimentController, selectedUsers, addSshKeysCheckBox.isSelected()));
            dialog.hide();
        }
    }

    private static final class SelectableUser {

        private final GeniUrn urn;
        private final BooleanProperty selected;
        private final boolean fixed;

        public SelectableUser(GeniUrn urn, boolean selected, boolean fixed) {
            this.fixed = fixed;
            this.selected = new SimpleBooleanProperty(selected);
            this.urn = urn;
        }

        public boolean isFixed() {
            return fixed;
        }

        public boolean getSelected() {
            return selected.get();
        }

        public BooleanProperty selectedProperty() {
            return selected;
        }

        public void setSelected(boolean selected) {
            this.selected.set(selected);
        }

        public GeniUrn getUrn() {
            return urn;
        }

        @Override
        public String toString() {
            return urn.getResourceName();
        }
    }
}
