package be.iminds.ilabt.jfed.experimenter_gui.editor.channel_properties;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.CMIManager;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.cmi.ComponentManagerInfo;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.JFedConnectivityException;
import be.iminds.ilabt.jfed.rspec.model.Channel;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecChannel;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.concurrent.WorkerStateEvent;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ComboBox;
import javafx.scene.control.ProgressBar;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;

import java.util.Comparator;

/**
 * User: twalcari
 * Date: 4/23/2015
 * Time: 16:46
 */
public class ChannelPropertiesDialog extends BorderPane {

    private final FXRspecChannel channel;
    private final FXModelRspec modelRspec;

    private final JFedConfiguration config;
    private final AuthorityList authorityList;
    private final CMIManager cmiManager;


    private final ObjectProperty<ComponentManagerInfo.Resources.RefreshAdvertisementTask> activeRefreshTask =
            new SimpleObjectProperty<>(null);

    @FXML
    private ComboBox<ComponentManagerInfo> componentManagerInfoComboBox;

    @FXML
    private ComboBox<Channel> channelComboBox;

    @FXML
    private ProgressBar advertisementFetchProgressBar;
    @FXML
    private VBox advertisementFetchOverlay;

    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    private boolean success = false;

    ChannelPropertiesDialog(FXRspecChannel channel, FXModelRspec modelRspec, JFedConfiguration config,
                            AuthorityList authorityList, CMIManager cmiManager) {
        this.channel = channel;
        this.modelRspec = modelRspec;

        this.config = config;
        this.authorityList = authorityList;
        this.cmiManager = cmiManager;


        //initialize the Scene
        FXMLUtil.injectFXML(this);

        for (JFedConfiguration.TestbedDescription testbedDescription : config.getTestbedDescriptions()) {
            if (testbedDescription.hasChannelSupport()) {
                componentManagerInfoComboBox.getItems().add(cmiManager.getComponentManagerInfo(testbedDescription.getUrn()));
            }
        }


        componentManagerInfoComboBox.setConverter(StringConverters.CMI_STRING_CONVERTER);
        channelComboBox.setConverter(StringConverters.CHANNEL_CONVERTER);

        componentManagerInfoComboBox.getSelectionModel().selectedItemProperty().addListener((obj) -> updateChannels());


        advertisementFetchOverlay.visibleProperty().bind(activeRefreshTask.isNotNull());
        activeRefreshTask.addListener((observable, oldValue, newValue) -> {
            if (oldValue != null) {
                advertisementFetchProgressBar.progressProperty().unbind();
            }

            if (newValue != null) {
                advertisementFetchProgressBar.progressProperty().bind(newValue.progressProperty());
            }
        });

        //load values from the provided channel

        if (channel.getComponentManagerId() != null) {
            ComponentManagerInfo currentCmi = cmiManager.getComponentManagerInfo(channel.getComponentManagerId());
            componentManagerInfoComboBox.setValue(currentCmi);
        } else {
            //try to find the most probable CMI that will be used
            componentManagerInfoComboBox.getItems().sort(
                    Comparator.comparing(this::getNumberOfNodesInRspec).reversed().thenComparing(cmi -> cmi.getAuthority().getName())
            );

            componentManagerInfoComboBox.getSelectionModel().selectFirst();
        }

        if (channel.getComponentId() != null) {
            channelComboBox.setValue(
                    channelComboBox.getItems().stream()
                            .filter(ch -> ch.getComponentId().equals(channel.getComponentId()))
                            .findAny().orElse(channel)
            );
        }

    }

    private int getNumberOfNodesInRspec(ComponentManagerInfo cmi) {
        return modelRspec.getNodes()
                .filtered(node -> node.getComponentManagerId() != null &&
                        node.getComponentManagerId().equals(cmi.getAuthority().getUrn()))
                .size();
    }

    private void updateChannels() {
        ComponentManagerInfo componentManagerInfo = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();

        if (componentManagerInfo != null) {
            //do we have the manifest available?
            ComponentManagerInfo.Resources resources = componentManagerInfo.getAllResources();
            if (resources.getAdvertisementRspecSource() == null) {
                resources = componentManagerInfo.getAvailableResources();
            }

            if (resources.getAdvertisementRspecSource() == null) {
                onRefreshAdvertisementAction();
            } else {
                channelComboBox.getItems().setAll(resources.getAdvertisementRspecSource().getModelRspec(ModelRspecType.FX).getChannels());
            }
        }
    }

    @FXML
    private void onRefreshAdvertisementAction() {
        ComponentManagerInfo componentManagerInfo = componentManagerInfoComboBox.getSelectionModel().getSelectedItem();
        assert (componentManagerInfo != null);

        final ComponentManagerInfo.Resources.RefreshAdvertisementTask updateAdvTask =
                componentManagerInfo.getAllResources().refreshAdvertisement();

        activeRefreshTask.set(updateAdvTask);

        final EventHandler<WorkerStateEvent> removeOnFinished = value -> {
            if (activeRefreshTask.get() == updateAdvTask) {
                activeRefreshTask.set(null);
            }

            updateChannels();
        };
        updateAdvTask.setOnSucceeded(removeOnFinished);
        updateAdvTask.setOnCancelled(removeOnFinished);

        updateAdvTask.setOnFailed(workerStateEvent -> {
            removeOnFinished.handle(workerStateEvent);

            //only show messages if pane is still visible
            if (ChannelPropertiesDialog.this.getScene() != null) {
                if (isConnectivityException(updateAdvTask.getException())) {
                    JFDialogs.create()
                            .owner(ChannelPropertiesDialog.this)
                            .message(String.format("A connectivity exception occurred while fetching the advertisement RSpec from %s.\n\n" +
                                            "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.",
                                    componentManagerInfo.getAuthority().getName()))
                            .masthead("Fetching advertisement RSpec failed")
                            .title("Fetching advertisement RSpec failed")
                            .showException(updateAdvTask.getException());


                } else {
                    JFDialogs.create()
                            .owner(ChannelPropertiesDialog.this)
                            .message(String.format("An error occurred while fetching the advertisement RSpec from %s.",
                                    componentManagerInfo.getAuthority().getName()))
                            .masthead("Fetching advertisement RSpec failed")
                            .title("Fetching advertisement RSpec failed")
                            .showException(updateAdvTask.getException());

                }
            }
        });

    }

    private static boolean isConnectivityException(Throwable t) {

        while (t != null) {
            if (t.getCause() instanceof JFedConnectivityException)
                return true;
            else
                t = t.getCause();
        }
        return false;
    }


    @FXML
    public void onSaveAction(ActionEvent actionEvent) {

        if (channelComboBox.getValue() == null) {
            JFDialogs.create()
                    .owner(this)
                    .message("Please select a channel from the dropdown list.")
                    .masthead("No channel selected.")
                    .showError();
            return;
        }

        Channel selectedChannel = channelComboBox.getValue();
        channel.setComponentManagerId(selectedChannel.getComponentManagerId());
        channel.setComponentId(selectedChannel.getComponentId());
        channel.setComponentName(selectedChannel.getComponentName());

        success = true;

        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();

    }


    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }

    public boolean isSuccess() {
        return success;
    }
}
