package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.experimenter_gui.util.ExecuteOnNotNull;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.beans.InvalidationListener;
import javafx.beans.WeakInvalidationListener;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableStringValue;
import javafx.collections.ListChangeListener;
import javafx.collections.WeakListChangeListener;

import java.time.Instant;
import java.time.ZoneId;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 14:58
 */
public class ExperimentStatusProperty implements ObservableStringValue {

    private final StringProperty status = new SimpleStringProperty();
    private final Experiment experiment;

    private final Timeline expirationTickerTimeline;
    private final InvalidationListener expirationDateInvalidationListener = observable -> updateExpirationStatus();

    private final ListChangeListener<Sliver> sliverListChangeListener = c -> {
        while (c.next()) {
            if (c.wasAdded()) {
                c.getAddedSubList().forEach(
                        sliver -> sliver.expirationDateProperty().addListener(
                                new WeakInvalidationListener(expirationDateInvalidationListener)));
            }
        }
        updateExpirationStatus();
    };

    public ExperimentStatusProperty(Experiment experiment) {
        this.experiment = experiment;

        this.expirationTickerTimeline = new Timeline(
                new KeyFrame(javafx.util.Duration.ZERO, actionEvent -> updateStatusString()),
                new KeyFrame(new javafx.util.Duration(1000))
        );
        this.expirationTickerTimeline.setCycleCount(Animation.INDEFINITE);

        new ExecuteOnNotNull<Slice>(experiment.sliceProperty()) {
            @Override
            public void run() {
                registerSlice();
            }
        };
    }

    private void registerSlice() {
        experiment.getSlice().expirationDateProperty().addListener(
                new WeakInvalidationListener(expirationDateInvalidationListener));

        experiment.getSlice().getSlivers().addListener(new WeakListChangeListener<>(sliverListChangeListener));
        experiment.getSlice().getSlivers().forEach(
                sliver -> sliver.expirationDateProperty().addListener(
                        new WeakInvalidationListener(expirationDateInvalidationListener)));

        updateExpirationStatus();
    }


    private void updateStatusString() {
        if (experiment.getRequestedStartTime() != null && experiment.getRequestedStartTime().isAfter(Instant.now())) {
            status.set(composeReservationStartMessage());
        } else {
            status.set(composeExpirationMessage());
        }
    }

    private String composeReservationStartMessage() {
        assert (experiment.getRequestedStartTime() != null && !experiment.getRequestedStartTime().isBefore(Instant.now()));


        java.time.Duration duration =
                java.time.Duration.between(Instant.now(), experiment.getRequestedStartTime()).withNanos(0);

        return String.format("The reservation for this experiment run will start on %s. It will start in %s.",
                DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(experiment.getRequestedStartTime().atZone(ZoneId.systemDefault())),
                DateTimeUtils.prettyPrintDuration(duration));

    }

    private String composeExpirationMessage() {
        if(nextExperimentExpiration.getFirstExpirationTime() == null)
            return "";

        StringBuilder message = new StringBuilder();

        if (nextExperimentExpiration.getFirstExpiringAuthority()  == null) {
            message.append("This experiment run");
        } else {
            message.append("The resources on ").append(nextExperimentExpiration.getFirstExpiringAuthority().getHrn());
        }


        if (nextExperimentExpiration.getFirstExpirationTime().isAfter(Instant.now())) {
            java.time.Duration duration = java.time.Duration.between(Instant.now(),
                    nextExperimentExpiration.getFirstExpirationTime()).withNanos(0);


            message.append(" will expire in ")
                    .append(DateTimeUtils.prettyPrintDuration(duration))
                    .append(".");
        } else {
            if (nextExperimentExpiration.getFirstExpiringAuthority() == null)
                message.append(" has EXPIRED!");
            else {
                message.append(" have EXPIRED!");
            }
        }

        return message.toString();
    }

    private NextExperimentExpiration nextExperimentExpiration;

    private void updateExpirationStatus() {
        assert (experiment.getSlice() != null);

        this.nextExperimentExpiration = new NextExperimentExpiration(experiment.getSlice());
        updateStatusString();

        expirationTickerTimeline.play();
    }

    @Override
    public String get() {
        return status.get();
    }

    @Override
    public void addListener(ChangeListener<? super String> listener) {
        status.addListener(listener);
    }

    @Override
    public void removeListener(ChangeListener<? super String> listener) {
        status.removeListener(listener);
    }

    @Override
    public String getValue() {
        return status.getValue();
    }

    @Override
    public void addListener(InvalidationListener listener) {
        status.addListener(listener);
    }

    @Override
    public void removeListener(InvalidationListener listener) {
        status.removeListener(listener);
    }
}
