package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.call_gui.TasksWindow;
import be.iminds.ilabt.jfed.experimenter_gui.controller.view.ControllerUI;
import be.iminds.ilabt.jfed.experimenter_gui.slice.errors.ErrorsView;
import be.iminds.ilabt.jfed.experimenter_gui.slice.errors.ErrorsViewTab;
import be.iminds.ilabt.jfed.experimenter_gui.slice.events.RenewExperimentEvent;
import be.iminds.ilabt.jfed.experimenter_gui.slice.progress.ProgressItem;
import be.iminds.ilabt.jfed.experimenter_gui.slice.progress.ProgressView;
import be.iminds.ilabt.jfed.experimenter_gui.slice.progress.ProgressViewTab;
import be.iminds.ilabt.jfed.experimenter_gui.slice.raw.RawSliceView;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.CustomDetailsView;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTask;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.SingleCallExperimentTask;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecMerger;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.eventbus.EventBus;
import javafx.application.Platform;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.WeakInvalidationListener;
import javafx.beans.value.ObservableStringValue;
import javafx.geometry.Orientation;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.Pane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.time.Duration;
import java.time.Instant;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 14:06
 */
public class ExperimentViewController extends SplitPane implements ExperimentChangeListener, ExperimentControllerListener, DetailsTabHost {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentViewController.class);

    private final TasksWindow tasksWindow;
    private final JFedGuiPreferences jFedPreferences;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final EventBus eventBus;

    /*
    The different main views
     */
    private final ModelSliceView modelSliceView;
    private final RawSliceView rawSliceView;
    private final ControllerUI controllerUi;

    /*
    Details TabPane + the default tabs
    */
    private final TabPane detailsTabPane;

    private final ProgressView progressView;
    private final ErrorsView errorsView;

    private final ExperimentStatusProperty experimentStatusProperty;
    private final ExpirationNotifier expirationNotifier;

    private final ExperimentController experimentController;

    /*
    The variables which represent the state that must be visualised
     */
    private final Experiment experiment;
    /**
     * This is the request rspec that is used to visualise the experiment in the topology view.
     * This differs from the normal requestRspecSource in that it also contains all the nodes from the already active
     * slivers in the slice.
     */
    private final ManifestRspecSource canvasRspecSource;


    private final InvalidationListener sceneInvalidationListener = new InvalidationListener() {
        @Override
        public void invalidated(Observable observable) {
            if (getScene() != null)
                experimentController.setParentWindow(getScene().getWindow());
            else
                experimentController.setParentWindow(null);
        }
    };

    public ExperimentViewController(ExperimentController experimentController, ExperimentViewsFactory experimentViewsFactory,
                                    TasksWindow tasksWindow, JFedGuiPreferences jFedPreferences,
                                    ExperimentTasksFactory experimentTasksFactory, EventBus eventBus) {
        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.experiment = experimentController.getExperiment();

        this.experimentStatusProperty = new ExperimentStatusProperty(experiment);
        this.expirationNotifier = new ExpirationNotifier(this, jFedPreferences, eventBus);
        this.eventBus = eventBus;

        this.tasksWindow = tasksWindow;
        this.jFedPreferences = jFedPreferences;

        experimentController.addListener(this);
        experiment.addExperimentChangeListener(this);

        this.canvasRspecSource =
                mergeNewRequestRspecWithExistingManifestRspec(experiment.getNewRequestRspecSource(), experiment.getSlice());

        sceneProperty().addListener(new WeakInvalidationListener(sceneInvalidationListener));

        //initialize detailsTabPane. This must be done BEFORE initializing the ControllerUI, as it adds tabs to this detailsPane
        detailsTabPane = new TabPane();
        detailsTabPane.setTabClosingPolicy(TabPane.TabClosingPolicy.ALL_TABS);

        //add progressview and errorsview to detailsTabPane
        this.progressView = new ProgressView();
        this.errorsView = experimentViewsFactory.createErrorsView();
        detailsTabPane.getTabs().addAll(new ProgressViewTab(progressView), new ErrorsViewTab(errorsView));

        //initialize the views
        this.modelSliceView = experimentViewsFactory.createModelSliceView(experimentController, canvasRspecSource);
        this.rawSliceView = experimentViewsFactory.createRawSliceView(experiment);
        this.controllerUi = experimentViewsFactory.createControllerUi(experiment, canvasRspecSource, this);

        ScrollPane detailsScrollPane = new ScrollPane();
        detailsScrollPane.setPrefHeight(100);
        detailsScrollPane.setContent(detailsTabPane);
        detailsScrollPane.setFitToHeight(true);
        detailsScrollPane.setFitToWidth(true);

        //configure SplitPane
        setOrientation(Orientation.VERTICAL);
        getItems().addAll(new Pane(), detailsScrollPane);
        setDividerPosition(0, 0.7);


        //now that the view has fully initialized and registered itself to the ExperimentController, we can start the latter
        experimentController.start();
    }


    @Override
    public void onSliceStateChange(SliceState newSliceState) {
        if (newSliceState == SliceState.CREATED) {
            Platform.runLater(this::checkAndWarnForExpiringResources);
        } else if (newSliceState == SliceState.PROVISIONING) {
            Platform.runLater(this::showSliversInitializationDialog);
        } else if (newSliceState == SliceState.RESERVATION_FAILED) {
            Platform.runLater(this::showReservationFailedDialog);
        } else if (newSliceState == SliceState.FAILED) {
            Platform.runLater(this::showFailedDialog);
        }
    }

    private void checkAndWarnForExpiringResources() {

        if (jFedPreferences.isRecoverExpirationWarningEnabled()
                &&         /*   only warn if we are recovering a slice! */
                experiment.getNewRequestRspecSource() == null) {

            NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment.getSlice());
            if (nextExperimentExpiration.getFirstExpirationTime() != null) {

                Duration durationToNextExpiration =
                        Duration.between(Instant.now(), nextExperimentExpiration.getFirstExpirationTime());
                if (durationToNextExpiration.toMinutes() < (long) jFedPreferences.getRecoverExpirationWarningTime()) {

                    LOG.debug("Showing warning dialog about imminent expiration of recovering experiment {}", experiment.getName());

                    ButtonType renewNewButtonType = new ButtonType("Renew now", ButtonBar.ButtonData.OK_DONE);
                    ButtonType ignoreButtonType = new ButtonType("Ignore", ButtonBar.ButtonData.CANCEL_CLOSE);


                    Alert alert =
                            new Alert(Alert.AlertType.WARNING,
                                    String.format("The experiment %s is expiring in less then %d minutes.\n" +
                                                    "Do you want to renew your experiment now?",
                                            experiment.getName(), durationToNextExpiration.toMinutes() + 1)
                                    , renewNewButtonType, ignoreButtonType);
                    alert.setHeaderText("Experiment is expiring soon");

                    alert.showAndWait().filter(buttonType -> buttonType == renewNewButtonType).ifPresent(buttonType -> {
                        eventBus.post(new RenewExperimentEvent(getScene().getWindow(), getExperimentController()));
                    });

                } else {

                    LOG.debug("Recovering experiment {} expires in {}, while treshold is on {} minutes. Not showing warning dialog.",
                            experiment.getName(), durationToNextExpiration.toMinutes(), jFedPreferences.getRecoverExpirationWarningTime());
                }
            }
        }
    }

    private void showReservationFailedDialog() {
        LOG.debug("Showing reservation failed dialog");


        StringBuilder content = new StringBuilder();
        content.append("The following resources could not be reserved:\n");

        for (GeniUrn geniUrn : experiment.getFailedReservationResources()) {
            content.append("\t- ").append(geniUrn.toString()).append("\n");
        }

        content.append("\nIt is advised to delete all requested resources from your experiment now.\n\nDo you want do delete all your resources?");


        DialogPane dialogPane = new DialogPane();
        dialogPane.setHeaderText("Reservation failed");
        dialogPane.setContentText(content.toString());
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.CALENDAR, Color.ORANGERED));

        dialogPane.getButtonTypes().setAll(ButtonType.YES, ButtonType.NO);


        Dialog<ButtonType> dialog = new Dialog<>();
        dialog.setTitle("Experiment run is initializing");
        dialog.setDialogPane(dialogPane);

        if (dialog.showAndWait().orElse(ButtonType.YES) == ButtonType.YES) {

            //terminate the experiment to actively delete all resources

            experimentController.submitExperimentTask(experimentTasksFactory.createStopExperimentTask(experiment));
        }

    }

    private void showFailedDialog() {
        LOG.debug("Showing failed dialog");


        DialogPane dialogPane = new DialogPane();
        dialogPane.setHeaderText("Experiment failed");
        dialogPane.setContentText("One or more resources in your experiment failed.\n\n" +
                "It is advised to delete all requested resources from your experiment now." +
                "\n\nDo you want do delete all your resources?");
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));

        dialogPane.getButtonTypes().setAll(ButtonType.YES, ButtonType.NO);


        Dialog<ButtonType> dialog = new Dialog<>();
        dialog.setTitle("Experiment run is initializing");
        dialog.setDialogPane(dialogPane);

        if (dialog.showAndWait().orElse(ButtonType.YES) == ButtonType.YES) {

            //terminate the experiment to actively delete all resources

            experimentController.submitExperimentTask(experimentTasksFactory.createStopExperimentTask(experiment));
        }

    }


    private static ManifestRspecSource mergeNewRequestRspecWithExistingManifestRspec(RequestRspecSource newRequestRspecSource, Slice slice) {
        ManifestRspecMerger manifestRspecMerger = new ManifestRspecMerger(ModelRspecType.FX);
        if (newRequestRspecSource != null)
            manifestRspecMerger.setRequest(newRequestRspecSource);
        else
            manifestRspecMerger.setRequest(slice.getRequestRspec());
        manifestRspecMerger.setAddNodesOnlyFoundInRequest(true);

        if (slice != null) {
            for (Sliver sliver : slice.getSlivers()) {
                manifestRspecMerger.setManifest(sliver.getAuthority().getUrn(), sliver.getManifestRspec());
            }
        }

        return manifestRspecMerger.getMergedManifest();
    }

    @Override
    public void addDetailsTab(Tab tab) {
        detailsTabPane.getTabs().add(tab);
    }

    @Override
    public void removeDetailsTab(Tab tab) {
        detailsTabPane.getTabs().remove(tab);
    }

    @Override
    public boolean containsDetailsTab(Tab tab) {
        return detailsTabPane.getTabs().contains(tab);
    }

    @Override
    public void selectDetailsTab(Tab tab) {
        assert (detailsTabPane.getTabs().contains(tab));
        detailsTabPane.getSelectionModel().select(tab);
    }

    private Node getMainView() {
        return getItems().get(0);
    }

    public boolean switchToModelView() {
        setMainView(modelSliceView);
        return true;

    }

    public boolean switchToRawView() {
        setMainView(rawSliceView);
        return true;

    }

    public boolean switchToExperimentController() {
        setMainView(controllerUi);
        return true;
    }

    /**
     * This method sets the element in the upper part of the splitpane
     *
     * @param node
     */
    private void setMainView(@Nonnull Node node) {
        Node oldNode = getItems().isEmpty() ? null : getItems().get(0);

        double[] oldDividerPostions = getDividerPositions();
        getItems().set(0, node);

        node.setVisible(true);
        if (oldNode != null)
            oldNode.setVisible(false);

        //FEDIBBTDEV-1006: force restoration of old  divider positions
        setDividerPositions(oldDividerPostions);
    }

    /*
    public ModelSliceView getModelSliceView() {
        return modelSliceView;
    }     */


    public ControllerUI getControllerUi() {
        return controllerUi;
    }

    public boolean isCanvasVisible() {
        return getMainView() == modelSliceView;
    }

    public boolean isControllerVisible() {
        return getMainView() == controllerUi;
    }

    public boolean isRawRspecVisible() {
        return getMainView() == rawSliceView;
    }


    /**
     * Shows a mutable dialog with information about the slice initialization
     */
    public void showSliversInitializationDialog() {


        //show a (mutable) notification to mac-users that the terminal opens in the background
        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SHOW_SLICE_INITIALIZATION_DIALOG, true)) {
            LOG.debug("Showing slivers initialization dialog");

            VBox pane = new VBox();
            pane.setPrefWidth(360);
            pane.setPrefHeight(250);
            pane.setSpacing(5);

            Label textLabel = new Label(
                    "It will take some time to start and initialize all nodes. " +
                            "You will be able to login by double-clicking or right-clicking on the node when it becomes green. " +
                            "Typically it takes about 3 minutes. " +
                            "Planetlab nodes take about 10-15 minutes before you can login through ssh (after they become green).\n\n" +
                            "If SSH-login does not work when the node is green, try again after some time.");
            textLabel.setPrefHeight(180); //work around bug FEDIBBTDEV-1008
            textLabel.setWrapText(true);
            pane.getChildren().add(textLabel);
            final CheckBox checkBox = new CheckBox("Do not show this dialog again");
            pane.getChildren().add(checkBox);

            DialogPane dialogPane = new DialogPane();
            dialogPane.setHeaderText("Your experiment run is initializing...");
            dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));
            dialogPane.setContent(pane);
            dialogPane.getButtonTypes().setAll(ButtonType.OK);


            Dialog<ButtonType> dialog = new Dialog<>();
            dialog.setTitle("Experiment run is initializing");
            dialog.setDialogPane(dialogPane);

            dialog.showAndWait();

            jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SHOW_SLICE_INITIALIZATION_DIALOG,
                    !checkBox.isSelected());
        } else {
            LOG.debug("Skipping slivers initializing dialog because user muted it before");
        }
    }


    public ExperimentController getExperimentController() {
        return experimentController;
    }

    @Override
    public void onExperimentTaskSubmitted(ExperimentTask experimentTask) {
        Platform.runLater(() -> {
            ProgressItem progressItem = experimentTask.getProgressItem();
            if (experimentTask instanceof CustomDetailsView) {
                CustomDetailsView customDetailsView = (CustomDetailsView) experimentTask;
                progressItem.setOnDetailsView(event -> {
                    if (customDetailsView.getCustomDetailsView() != null) {
                        customDetailsView.getCustomDetailsView().handle(event);
                    }
                });
            } else if (experimentTask instanceof SingleCallExperimentTask) {
                SingleCallExperimentTask singleCallExperimentTask = (SingleCallExperimentTask) experimentTask;
                progressItem.setOnDetailsView(event -> {
                    if (singleCallExperimentTask.getTaskExecution() != null) {
                        tasksWindow.showTaskExecution(singleCallExperimentTask.getTaskExecution());
                    }
                });
            }

            progressView.registerProgressViewItem(progressItem);


            experimentTask.addTaskExecutionsListener(errorsView);
        });

    }

    public Experiment getExperiment() {
        return experimentController.getExperiment();
    }

    public ObservableStringValue statusProperty() {
        return experimentStatusProperty;
    }

    public ModelSliceView getModelSliceView() {
        return modelSliceView;
    }

    public JFedGuiPreferences getjFedPreferences() {
        return jFedPreferences;
    }
}
