package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;

import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/2/2015
 * Time: 17:01
 */
public class RenewSliversAtAuthoritiesTask extends ExperimentTask {

    private final Instant newExpirationDate;
    private final Collection<SfaAuthority> authorities;
    private final HighLevelController hlc;

    RenewSliversAtAuthoritiesTask(Experiment experiment, Instant newExpirationDate, Collection<SfaAuthority> authorities, HighLevelController hlc) {
        super(experiment);
        this.newExpirationDate = newExpirationDate;
        this.authorities = authorities;
        this.hlc = hlc;

        updateTitle("Renew experiment until " + DateTimeFormatter.RFC_1123_DATE_TIME.format(ZonedDateTime.ofInstant(newExpirationDate, ZoneId.systemDefault())));
    }


    @Override
    protected TaskStatusIndicator.Status executeTask() {

        if (experiment.getSlice().getExpirationDate() == null ||
                newExpirationDate.isAfter(experiment.getSlice().getExpirationDate())) {

            updateMessage("Renewing experiment container");

            HighLevelController.RenewSliceTask renewSliceTask =
                    hlc.renewSlice(experiment.getSlice(), newExpirationDate, false/*renewAllSlivers*/);

            TaskExecution renewSliceTaskExecution = hlc.submitTask(renewSliceTask);
            waitForTaskExecutionToComplete(renewSliceTask, renewSliceTaskExecution);

            if (renewSliceTaskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                return TaskStatusIndicator.Status.FAILED;
            }

        }

        List<SfaAuthority> failedAuthorities = new ArrayList<>();

        for (SfaAuthority authority : authorities) {
            if (authority.isFake())
                continue;

            Task renewSliversTask =
                    hlc.renewSliversAtAuthority(experiment.getSlice(), newExpirationDate, authority);

            updateMessage("Renewing at %s");

            TaskExecution renewSliversTaskExecution = hlc.submitTask(renewSliversTask);
            waitForTaskExecutionToComplete(renewSliversTask, renewSliversTaskExecution);

            if (renewSliversTaskExecution.getState() != TaskExecution.TaskState.SUCCESS)
                failedAuthorities.add(authority);
        }


        if (failedAuthorities.isEmpty()) {
            updateMessage("");
            return TaskStatusIndicator.Status.SUCCESS;
        } else {
            updateMessage("Renewing failed at " + failedAuthorities.stream().map(SfaAuthority::getHrn).collect(Collectors.joining(", ")));

            return failedAuthorities.size() == authorities.size() ?
                    TaskStatusIndicator.Status.FAILED
                    : TaskStatusIndicator.Status.WARNING;
        }

    }

    @Override
    public String toString() {
        return "RenewSliversAtAuthoritiesTask{" +
                "newExpirationDate=" + newExpirationDate +
                ", authorities=" + authorities +
                '}';
    }
}
