package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 12:54
 */
public class RestoreSliceAndSliversTaskGenerator implements Runnable {
    private static final Logger LOG = LoggerFactory.getLogger(RestoreSliceAndSliversTaskGenerator.class);

    private final ExperimentController experimentController;
    private final Experiment experiment;
    private final HighLevelController hlc;
    private final AuthorityListModel authorityListModel;
    private final AuthorityFinder authorityFinder;

    RestoreSliceAndSliversTaskGenerator(ExperimentController experimentController, HighLevelController hlc, AuthorityListModel authorityListModel, AuthorityFinder authorityFinder) {
        this.experimentController = experimentController;
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.experiment = experimentController.getExperiment();
        this.hlc = hlc;
    }


    @Override
    public void run() {
        assert (experiment.getSlice() != null);

        if (!hasSufficientSliceInformation()) {
            //we have insufficient information to continue. We need to fetch it first!
            requestSliceInformation();
        } else {
            restoreSlivers();
        }


    }

    private boolean hasSufficientSliceInformation() {
        return (experiment.getSlice().getExpirationDate() != null);
    }

    private void requestSliceInformation() {
        LOG.info("Requesting slice information for {}", experiment.getSlice().getUrn());
        experimentController.submitExperimentTask(new ResolveSliceExperimentTask(experiment, hlc));
    }

    private CountDownLatch allAuthoritiesContactedLatch = null;

    private void restoreSlivers() {
        assert hasSufficientSliceInformation();

        if (experiment.getExistingRequestRspecSource() != null) {
            assert experiment.getExistingRequestRspecSource().getStringRspec().isWellFormed();

            //request a status update from each authority which is currently present in the manifest
            List<SfaAuthority> authoritiesToContact =
                    experiment.getExistingRequestRspecSource().getAllComponentManagerUrns().stream()
                            .map(urn -> authorityFinder.findByUrn(urn, AuthorityFinder.Purpose.RECOVER))
                                    .filter(Objects::nonNull)  //should not happen..
                                    .filter(SfaAuthority::isReal)
                                    .collect(Collectors.toList());


            if (!authoritiesToContact.isEmpty()) {
                LOG.debug("Going to contact authorities {} to recover slice", authoritiesToContact);
                allAuthoritiesContactedLatch = new CountDownLatch(authoritiesToContact.size());


                authoritiesToContact.forEach(auth -> experimentController.submitExperimentTask(new RestoreManifestExperimentTask(experiment, auth, hlc)));
                //wait for all authorities to be contacted
                try {
                    allAuthoritiesContactedLatch.await();
                } catch (InterruptedException ignored) {
                }
            } else {
                LOG.debug("No additional authorities to contact");
            }
        } else {
            LOG.info("No (additional) slivers to restore as there is no existingRequestRspecSource for this experiment!");
        }

        //ready!
        experiment.setSliceState(SliceState.CREATED);

    }


    private final class ResolveSliceExperimentTask extends SingleCallExperimentTask {
        public ResolveSliceExperimentTask(Experiment experiment, HighLevelController hlc) {
            super(experiment, hlc);

            updateTitle("Requesting slice information");
        }

        @Override
        public Task createTask() {
            return hlc.getAggregatesForSlice(experiment.getSlice());
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (taskExecution.getState() == TaskExecution.TaskState.SUCCESS) {
                assert (experiment.getSlice().getExpirationDate() != null);

                if (experiment.getSlice().getExpirationDate().isBefore(Instant.now())) {
                    updateMessage("Experiment has already expired. Cannot continue.");
                    LOG.error("Slice has already expired on {}, cannot continue.", experiment.getSlice().getExpirationDate());

                    experiment.setSliceState(SliceState.EXPIRED_OR_EMPTY);
                } else {
                    restoreSlivers();
                }
            }
        }
    }

    private final class RestoreManifestExperimentTask extends SingleCallExperimentTask {

        private final SfaAuthority authority;

        public RestoreManifestExperimentTask(Experiment experiment, SfaAuthority authority, HighLevelController hlc) {
            super(experiment, hlc);
            this.authority = authority;

            updateTitle("Restore manifest from " + authority.getName());
        }

        @Override
        public Task createTask() {
            return hlc.getSliceManifest(experiment.getSlice(), authority);
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (taskExecution.getState() == TaskExecution.TaskState.SUCCESS) {

                if (experiment.getRequestedStartTime() != null && experiment.getRequestedStartTime().isAfter(Instant.now())) {
                    //do nothing
                    allAuthoritiesContactedLatch.countDown();
                } else {
                    experimentController.submitExperimentTask(new SafeStatusUntilReadyTask(experiment, authority, hlc));
                }
            } else {
                allAuthoritiesContactedLatch.countDown();
            }
        }
    }

    private final class SafeStatusUntilReadyTask extends StatusUntilReadyTask {

        public SafeStatusUntilReadyTask(Experiment experiment, SfaAuthority authority, HighLevelController hlc) {
            super(experiment, authority, hlc);
        }

        @Override
        protected void onFinished() {
            if (getStatusUntilReadyOrFailTask.getLastCompletedTaskExecution().getState() == TaskExecution.TaskState.FAILED) {
                //assume that the sliver is not longer allocated
                experiment.getSlice().getSlivers().stream()
                        .filter(sliver -> sliver.getAuthority() == authority)
                        .forEach(sliver -> {
                            sliver.setStatus(new StatusDetails(StatusDetails.SliverStatus.UNALLOCATED));
                        });
            }

            allAuthoritiesContactedLatch.countDown();
        }
    }
}
