package be.iminds.ilabt.jfed.gts;

import be.iminds.ilabt.jfed.groovy_dsl.DSLEventFactory;
import be.iminds.ilabt.jfed.groovy_dsl.DSLEventWriter;
import be.iminds.ilabt.jfed.groovy_dsl.DSLOutputFactory;
import be.iminds.ilabt.jfed.groovy_dsl.DSLStreamException;
import be.iminds.ilabt.jfed.groovy_dsl.events.DSLEvent;
import be.iminds.ilabt.jfed.groovy_dsl.helpers.IndentingDSLEventWriter;
import be.iminds.ilabt.jfed.groovy_dsl.impl.events.ClosureRefArgumentImpl;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.Callable;

import static be.iminds.ilabt.jfed.gts.GtsDslConstants.*;

/**
 * User: twalcari
 * Date: 5/18/2015
 * Time: 10:50
 */
public class GtsDslWriter implements Callable<String> {
    private static final Logger LOG = LoggerFactory.getLogger(GtsDslWriter.class);

    private final String name;
    private final ModelRspec model;

    public GtsDslWriter(String name, ModelRspec model) {
        this.name = name;
        this.model = model;
    }


    @Override
    public String call() throws DSLStreamException{

        DSLOutputFactory outputFactory = DSLOutputFactory.newInstance();
        StringWriter sw = new StringWriter();

        DSLEventWriter eventWriter = outputFactory.createDSLEventWriter(sw);
        eventWriter = new IndentingDSLEventWriter(eventWriter);

        DSLEventFactory eventFactory = DSLEventFactory.newInstance();


        eventWriter.add(eventFactory.createClosureStartElement(name));


        for (RspecNode rspecNode : model.getNodes()) {

            if (!rspecNode.getComponentManagerId().equals(GTS_URN_PORT_CM)) {

                writeHost(rspecNode, eventWriter, eventFactory);
            } else {
                writePort(rspecNode, eventWriter, eventFactory);
            }
        }

        for (RspecLink rspecLink : model.getLinks()) {
            List<DSLEvent> adjacencyEvents = new ArrayList<>();
            eventWriter.add(eventFactory.createClosureStartElement(GTS_LINK));

            eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_ID, rspecLink.getClientId()));

            int ifaceNumber = 0;
            for (RspecInterface rspecInterface : rspecLink.getInterfaces()) {
                String ifaceName = "p" + (++ifaceNumber);

                eventWriter.add(eventFactory.createClosureStartElement(GTS_PORT));
                eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_ID, ifaceName));

                eventWriter.add(eventFactory.createClosureEndElement()); //GTS_PORT


                adjacencyEvents.add(eventFactory.createFunctionCallElement(GTS_FUNC_ADJACENCY,
                        new ClosureRefArgumentImpl(rspecInterface.getNode().getClientId(),
                                convertIfaceNameToPortName(rspecInterface.getClientId())),
                        new ClosureRefArgumentImpl(rspecLink.getClientId(), ifaceName)));

            }
            eventWriter.add(eventFactory.createClosureEndElement()); //GTS_LINK

            //now write all the adjacency's
            for (DSLEvent adjacencyEvent : adjacencyEvents) {
                eventWriter.add(adjacencyEvent);
            }
        }


        eventWriter.add(eventFactory.createClosureEndElement()); //name

        return sw.toString();
    }

    private void writePort(RspecNode rspecNode, DSLEventWriter eventWriter, DSLEventFactory eventFactory) throws DSLStreamException {
        assert (GTS_URN_PORT_CM.equals(rspecNode.getComponentManagerId()));

        eventWriter.add(eventFactory.createClosureStartElement(GTS_PORT));

        eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_ID, rspecNode.getClientId()));

        eventWriter.add(eventFactory.createClosureEndElement()); //GTS_PORT

    }

    private void writeHost(RspecNode rspecNode, DSLEventWriter eventWriter, DSLEventFactory eventFactory) throws DSLStreamException {
        assert (!GTS_URN_PORT_CM.equals(rspecNode.getComponentManagerId()));

        eventWriter.add(eventFactory.createClosureStartElement(GTS_HOST));
        eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_ID, rspecNode.getClientId()));

        if (rspecNode.getComponentId() != null) {
            boolean valid = true;
            if (!rspecNode.getComponentId().getTopLevelAuthority().equals(GTS_URN_TLA)) {
                LOG.warn("Unexpected TopLevelAuthority '{}' for componentId, only {} is supported! Ignoring.",
                        rspecNode.getComponentId().getTopLevelAuthority(), GTS_URN_TLA);
                valid = false;
            }

            if (!rspecNode.getComponentId().getResourceType().equals(GTS_URN_LOCATION_RESOURCE_TYPE)) {
                LOG.warn("Unexpected TopLevelAuthority '{}' for componentId, only {} is supported! Ignoring.",
                        rspecNode.getComponentId().getResourceType(), GTS_URN_LOCATION_RESOURCE_TYPE);
                valid = false;
            }

            if (valid) {
                eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_LOCATION, rspecNode.getComponentId().getResourceName()));
            }
        }

        Optional<? extends RspecInterface> controlIface = findControlInterface(rspecNode);

        if (controlIface.isPresent()) {
            RspecInterface iface = controlIface.get();

            eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_CONTROLLER_IPV4, iface.getPublicIpv4()));

        }


        for (RspecInterface rspecInterface : rspecNode.getInterfaces()) {
            eventWriter.add(eventFactory.createClosureStartElement(GTS_PORT));
            eventWriter.add(eventFactory.createStringPropertyElement(GTS_PROP_ID, convertIfaceNameToPortName(rspecInterface.getClientId())));


            eventWriter.add(eventFactory.createClosureEndElement()); //GTS_PORT
        }


        eventWriter.add(eventFactory.createClosureEndElement()); //GTS_HOST
    }


    private Optional<? extends RspecInterface> findControlInterface(RspecNode node) {

        return node.getInterfaces().stream()
                .filter(iface -> GTS_CONTROL_ROLE.equals(iface.getRole()))
                .findFirst();

    }

    private String convertIfaceNameToPortName(String ifaceName) {
        while (ifaceName.indexOf(':') >= 0) {
            ifaceName = ifaceName.substring(ifaceName.indexOf(':') + 1);
        }
        return ifaceName;
    }
}
