package be.iminds.ilabt.jfed.highlevel.stitcher;

import be.iminds.ilabt.jfed.call_log_output.CallReport;
import be.iminds.ilabt.jfed.call_log_output.HtmlLogOutput;
import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.highlevel.controller.*;
import be.iminds.ilabt.jfed.highlevel.model.*;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.StitchingComputationService;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.lowlevel.stitching.info.Hop;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import javafx.application.Platform;
import javafx.beans.binding.BooleanBinding;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * ParallelStitcher
 */
public class ParallelStitcher {
    private static final org.slf4j.Logger ACTUAL_LOG = LoggerFactory.getLogger(ParallelStitcher.class);
    private final StitchingReportWrappingLogger log = new StitchingReportWrappingLogger(ACTUAL_LOG, null/*report State*/);
    private final StitchingReportWrappingLogger stitchingDirectorLog = new StitchingReportWrappingLogger(LoggerFactory.getLogger(StitchingDirector.class), null/*report State*/);

    //can change for testing
    private final long TIME_BETWEEN_STATUS_CHECKS_MS;
    private final long STATUS_TIMEOUT_MS;
    private final long TIME_BEFORE_CREATESLIVER_RETRY_MS;

    public static enum GlobalPhase {SCS, ALLOCATE, PROVISION, WAIT_UNTIL_READY, SUCCESS, FAILED_TEMP, FAILED_PERMANENT}

    private GlobalPhase globalPhase = GlobalPhase.SCS;

    private CountDownLatch countdownLatch;


    private final StitchingReportFactory stitchingReportFactory;
    private final StitchingReport stitchingReport;

    private final static int MAX_CREATE_RETRIES = 3;
    private final static int MAX_STATUS_RETRIES = 3;


    private final ArrayList<TaskFinishedCallback> failedTaskCallbacks = new ArrayList<>();
    private final TaskFinishedCallback defaultTaskCallback = (task, taskExecution, state) -> {
        if (taskExecution.getState() == TaskExecution.TaskState.FAILED) {
            for (TaskFinishedCallback taskFinishedCallback : failedTaskCallbacks) {
                try {
                    taskFinishedCallback.onTaskExecutionFinished(task, taskExecution, state);
                } catch (Exception | AssertionError t) {
                    log.error("Exception in callback", t);
                    //ignore
                }
            }
        }
    };

    private final ReadOnlyListWrapper<StitchingActionTracker> sliverTrackers =
            new ReadOnlyListWrapper<>(FXCollections.synchronizedObservableList(
                    FXCollections.<StitchingActionTracker>observableArrayList()));

    private final ReadOnlyListWrapper<TaskWithExecution<SCSRequestTask>> scsRequestTasks =
            new ReadOnlyListWrapper<>(FXCollections.synchronizedObservableList(
                    FXCollections.<TaskWithExecution<SCSRequestTask>>observableArrayList()));

    private List<StitchingActionTracker> toProvision = Collections.synchronizedList(new ArrayList<>());
    private List<StitchingActionTracker> toWaitUntilReady = Collections.synchronizedList(new ArrayList<>());

    private final StitchingDirector director;
    private final HighLevelController highLevelController;
    private final GeniUserProvider geniUserProvider;
    private final JFedConnectionProvider connectionProvider;
    private final ObjectProperty<StitchingComputationService.ComputePathResult> scsResult =
            new SimpleObjectProperty<>();
    private final Map<SfaAuthority, StitchingActionTracker> trackers = Collections.synchronizedMap(new HashMap<>());
    private final SfaAuthority stitchingAuthority;
    private boolean started = false;
    private ReadOnlyBooleanWrapper scsCallFinished = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper scsCallOk = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper stitchingSuccess = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper stitchingFailed = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper deleteAllCalled = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper deleteAllSuccess = new ReadOnlyBooleanWrapper(false);
    private ReadOnlyBooleanWrapper deleteAllFailure = new ReadOnlyBooleanWrapper(false);
    private Slice slice;
    private List<AnyCredential> userCredentials;
    private RequestRspecSource originalRspecRequest;
    private final Logger logger;
    private GiveUpStitchingCallback callback = null;
    private final List<UserSpec> userSpecs;

    private final Instant requestedSliverExpirationDate;

    private final TaskThread taskThread;
    private final EasyModel easyModel;
    private final JFedPreferences jFedPreferences;
    private final AuthorityList authorityList;
    private final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
            automaticAggregateManagerWrapperFactory;

    public ParallelStitcher(TaskThread taskThread, EasyModel easyModel, GeniUserProvider geniUserProvider, RequestRspecSource originalRspecRequest,
                            JFedConnectionProvider connectionProvider, SfaAuthority stitchingAuthority, HighLevelController highLevelController,
                            List<UserSpec> userSpecs, Instant requestedSliverExpirationDate, AuthorityList authorityList, JFedPreferences jFedPreferences,
                            AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory,
                            Long timeBetweenStatusChecks, Long statusTimeoutMs, Long createSliverRetryDelayMs, Logger logger, StitchingReportFactory stitchingReportFactory) {

        assert authorityList != null;
        assert automaticAggregateManagerWrapperFactory != null;
        assert easyModel != null;
        assert taskThread != null;
        assert connectionProvider != null;
        assert highLevelController != null;
        assert geniUserProvider != null;
        assert originalRspecRequest != null;
        assert jFedPreferences != null;
        assert logger != null;
        assert stitchingAuthority != null;

        this.taskThread = taskThread;
        this.authorityList = authorityList;

        this.logger = logger;

        this.easyModel = easyModel;
        this.jFedPreferences = jFedPreferences;
        this.highLevelController = highLevelController;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;

        this.geniUserProvider = geniUserProvider;
        this.originalRspecRequest = originalRspecRequest;
        this.requestedSliverExpirationDate = requestedSliverExpirationDate;
        this.director = new StitchingDirector(
                this.authorityList.getAuthorityListModel(),
		this.authorityList.finder(),
                requestedSliverExpirationDate == null ? null : Date.from(requestedSliverExpirationDate),
                stitchingDirectorLog);
        this.connectionProvider = connectionProvider;
        this.stitchingAuthority = stitchingAuthority;
        this.userSpecs = userSpecs;

        this.TIME_BETWEEN_STATUS_CHECKS_MS = timeBetweenStatusChecks == null ? 10_000l : timeBetweenStatusChecks;
        this.STATUS_TIMEOUT_MS =
                statusTimeoutMs == null ? 20 * 60 * 1000 /*5 minutes to get to ready */ : statusTimeoutMs;
        this.TIME_BEFORE_CREATESLIVER_RETRY_MS = createSliverRetryDelayMs == null ? 10_000l : createSliverRetryDelayMs;

        this.stitchingReportFactory = stitchingReportFactory;
        this.stitchingReport = stitchingReportFactory.createStitchingReport("SCS");

        if (stitchingReport.getCurState() != null) {
            log.setState(stitchingReport.getCurState());
            stitchingDirectorLog.setState(stitchingReport.getCurState());
        }
    }

    public void addOnFailedTaskFinishedCallback(TaskFinishedCallback taskFinishedCallback) {
        failedTaskCallbacks.add(taskFinishedCallback);
    }

    public void removeOnFailedTaskFinishedCallback(TaskFinishedCallback taskFinishedCallback) {
        failedTaskCallbacks.remove(taskFinishedCallback);
    }


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////// interactive part of the interface ///////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * keep only SfaCredentials. Leave out ABAC.
     */
    protected static List<AnyCredential> filterCredentials(List<AnyCredential> origCredentials) {
        List<AnyCredential> res = new ArrayList<>();
        for (AnyCredential cred : origCredentials) {
            if (cred instanceof SfaCredential) {
                res.add(cred);
                return res; //only 1 credential needed
            }
        }
        if (res.isEmpty() && !origCredentials.isEmpty()) {
            ACTUAL_LOG.warn("Unknown what is a good credential to use. Will just use first one received (type={}).",
                    origCredentials.get(0).getClass().getName());
            res.add(origCredentials.get(0));
        }
        return res;
    }

    private AggregateManagerWrapper getAggregateManagerWrapper(TaskExecution taskExecution, SfaAuthority amAuth, StitchingReport.State state) {
        assert automaticAggregateManagerWrapperFactory != null;
        assert easyModel != null;
        return automaticAggregateManagerWrapperFactory.create(JavaFXLogger.wrap(easyModel.getLogger(), taskExecution, state), amAuth);
    }


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////// callback when stitching is given up ///////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Contact SCS and Create all slivers
     */
    public void start() {
        countdownLatch = new CountDownLatch(1);

        assert slice != null : "must set slice first";
        assert geniUserProvider.isUserLoggedIn() : "User needs to be logged in";

        if (started)
            throw new IllegalStateException("doScsRequest may only be called once!");
        started = true;

        log.debug("ParallelStitcher started.");
        doScsRequest();
    }

    public void setGiveUpStitchingCallback(GiveUpStitchingCallback c) {
        callback = c;
    }

    public void clearGiveUpStitchingCallback() {
        callback = null;
    }

    public void giveUp(String reason) {
        try {
            //report giveUp to director, in case director didn't give up but ParallelStitcher did.
            director.reportGiveup();

            CallReport.State state = stitchingReport.getCurState();
            state.setState(LogOutput.TestResultState.FAILED);
            changeState(GlobalPhase.FAILED_PERMANENT);

            boolean mustDeleteAll =
                    callback == null || callback.onGiveUpStitching(reason);
            if (mustDeleteAll) {
                // no need to ever register slivers, as we will delete them anyway
                log.info("Will delete all.");
                deleteAll();
            } else {
                log.info("Will not delete all: callback did not request it.");
                //we need to register the slivers
                registerAllSliverInfo(state, slice);
            }
        } finally {
            countdownLatch.countDown();
        }
    }

    public boolean isGaveUp() {
        return director.isGaveUp();
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////// read only feedback ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Abort all calls and call delete on all involved aggregates
     */
    public void deleteAll() {
        StitchingReport.State state = stitchingReport.getCurState();

        assert slice != null : "must set slice first";

        Platform.runLater(() -> deleteAllCalled.set(true));

        //stop all current tasks and calls
        for (final StitchingActionTracker tracker : sliverTrackers) {
            synchronized (tracker) {
                tracker.deletingAll = true;
                if (tracker.currentStatusTask != null)
                    tracker.currentStatusTask.cancel();

                tracker.createTasks.forEach(twe -> cancelOnTaskThread(twe.getTask()));
                tracker.statusTasks.forEach(twe -> cancelOnTaskThread(twe.getTask()));
            }
        }

        final IntegerProperty deleteAllCounter = new SimpleIntegerProperty(sliverTrackers.size());
        //call delete on all involved aggregates
        synchronized (sliverTrackers) {
            for (final StitchingActionTracker tracker : sliverTrackers) {
                if (tracker.privateStatus.equals(Status.INIT) || tracker.privateStatus.equals(Status.WAITING_FOR_DEP)) {
                    //no need to delete
                    log.debug("ParallelStitcher.delete() -> no need to Delete (state=" + tracker.privateStatus + ") on " + tracker.getAuth().getUrn());
                    continue;
                }
                log.debug("ParallelStitcher.delete() -> Deleting (state=" + tracker.privateStatus + ") on " + tracker.getAuth().getUrn());

                DeleteSliverTask deleteTask = deleteSliver(tracker, state);
                deleteTask.addCallback(defaultTaskCallback);
                deleteTask.addCallback((task, taskExecution, taskState) -> {
                    try {
                        if (taskState == TaskExecution.TaskState.SUCCESS) {
                            //todo handle delete success: set global delete status
                            tracker.setStatus(Status.INIT);
                            Platform.runLater(() -> {
                                int o = deleteAllCounter.get() - 1;
                                deleteAllCounter.set(o);
                                if (o == 0)
                                    deleteAllSuccess.set(true);
                            });
                        } else {
                            //todo handle delete failure: set global delete status
                            tracker.setStatus(Status.FAIL);
                            Platform.runLater(() -> {
                                deleteAllSuccess.set(false);
                                deleteAllFailure.set(true);
                            });
                        }
                    } catch (AssertionError t) {
                        log.error("exception in delete callback", t);
                        throw t;
                    } catch (Exception t) {
                        log.error("exception in delete callback", t);
                        throw t;
                    }
                });
                TaskExecution taskExecution = addTaskToTaskThread(deleteTask);

                Platform.runLater(() -> {
                    tracker.deleteTasks.add(new TaskWithExecution<>(deleteTask, taskExecution));
                });

            }
        }
    }

    public String getOriginalRspecRequestString() {
        return originalRspecRequest.getRspecXmlString();
    }

    public RequestRspecSource getOriginalRspecRequest() {
        return originalRspecRequest;
    }

    public ReadOnlyListProperty<StitchingActionTracker> getSliverTrackers() {
        return sliverTrackers.getReadOnlyProperty();
    }

    public Slice getSlice() {
        return slice;
    }

    public void setUserCredentialAndSlice(List<AnyCredential> userCredentials, Slice slice) {
        assert userCredentials != null;
        assert !userCredentials.isEmpty();
        this.userCredentials = userCredentials;
        assert slice != null;
        assert slice.hasCredential() : "Must use a slice with a known credential: " + slice.getUrnString();
        this.slice = slice;
    }

    public StitchingDirector getDirector() {
        return director;
    }

    public EasyModel getEasyModel() {
        return easyModel;
    }

    public ReadOnlyBooleanProperty stitchingSuccessProperty() {
        return stitchingSuccess.getReadOnlyProperty();
    }

    public ReadOnlyBooleanProperty stitchingFailedProperty() {
        return stitchingFailed.getReadOnlyProperty();
    }

    public BooleanBinding getStitchingFinished() {
        return stitchingFailed.or(stitchingSuccess);
    }

    public ReadOnlyListProperty<TaskWithExecution<SCSRequestTask>> getScsRequestTasks() {
        return scsRequestTasks.getReadOnlyProperty();
    }

    public boolean getStitchingSuccess() {
        return stitchingSuccess.get();
    }

    public boolean getStitchingFailed() {
        return stitchingFailed.get();
    }

    public boolean getDeleteAllCalled() {
        return deleteAllCalled.get();
    }

    public ReadOnlyBooleanProperty deleteAllCalledProperty() {
        return deleteAllCalled.getReadOnlyProperty();
    }

    public boolean getDeleteAllSuccess() {
        return deleteAllSuccess.get();
    }

    public ReadOnlyBooleanProperty deleteAllSuccessProperty() {
        return deleteAllSuccess.getReadOnlyProperty();
    }

    public boolean getDeleteAllFailure() {
        return deleteAllFailure.get();
    }

    public ReadOnlyBooleanProperty deleteAllFailureProperty() {
        return deleteAllFailure.getReadOnlyProperty();
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////// private internals /////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public ReadOnlyBooleanProperty scsCallOkProperty() {
        return scsCallOk.getReadOnlyProperty();
    }

    public ReadOnlyBooleanProperty scsCallFinishedProperty() {
        return scsCallFinished.getReadOnlyProperty();
    }

    public SfaAuthority getStitchingAuthority() {
        return stitchingAuthority;
    }

    private void doScsRequest() {
        final SCSRequestTask scsRequestTask = new SCSRequestTask(stitchingReport.getCurState());
        scsRequestTask.addCallback(defaultTaskCallback);
        TaskExecution taskExecution = addTaskToTaskThread(scsRequestTask);

        Platform.runLater(() -> {
            scsRequestTasks.add(new TaskWithExecution<>(scsRequestTask, taskExecution));
        });
    }

    /**
     * This method looks up the next slivers that can be initialized, and starts the creation.
     * It also checks if the global phase is done, and if a next phase needs to be started.
     */
    private void checkNextActionAndPhase() {
        log.debug("checkNextActionAndPhase()");
        StitchingReport.State state = stitchingReport.getCurState();
        log.debug("checkNextActionAndPhase() " + globalPhase + "   reportState=" + state.getName());

        try {
            Map<String, List<Hop>> hopOverview = director.getHopOverview();
            stitchingReport.updateHopOverview(hopOverview);
        } catch (Exception | AssertionError e) {
            log.warn("Error generating informative only stitching data for stitchingReport. This can safely be ignored without influencing the stitching itself.",e);
        }

        switch (globalPhase) {
            case SCS: {
                //waiting for SCS call
                break;
            }
            case ALLOCATE: {
                //allocating or creating slivers
                List<StitchingDirector.StitchAction> stichActions;
                boolean noMoreActionsNeeded;
                synchronized (director) {
                    stichActions = director.getStitchActions();

                    //now do actions requested. Note that we are still synchronized, otherwise things can get messy when
                    //  actions complete inside this loop, and states changes are not seen correctly
                    for (StitchingDirector.StitchAction stitchAction : stichActions) {
                        assert stitchAction.getAuthority() != null;
                        final StitchingActionTracker tracker = trackers.get(stitchAction.getAuthority());
                        assert tracker != null;
                        assert stitchAction.getAuthority().getUrnString().equals(tracker.getAuth().getUrnString());
                        assert stitchAction.getAuthority().equals(tracker.getAuth());

                        switch (stitchAction.getType()) {
                            case ALLOCATE: {
//                                if (tracker.privateStatus == Status.INIT || tracker.privateStatus == Status.WAITING_FOR_DEP) {
                                log.info("nextstep: ALLOCATE -> tracker (status=" + tracker.privateStatus + ") for " + stitchAction);
                                tracker.stitchAction = stitchAction;
                                tracker.setStatus(Status.WAITING_FOR_ALLOCATE);
                                createSliver(tracker, 0/*now*/, state);
//                                } else {
//                                    //it is already being created
//                                    log.info("ALREADY DOING nextstep: ALLOCATE -> tracker (status=" + tracker.privateStatus + ") for " + stitchAction);
//                                }
                                break;
                            }
                            case DELETE: {
//                                if (tracker.privateStatus != Status.FAIL) {
                                synchronized (tracker) {
                                    log.info("nextstep: DELETE -> tracker (status=" + tracker.privateStatus + ") for " + stitchAction);
                                    tracker.stitchAction = stitchAction;
                                    tracker.setStatus(Status.FAIL);
                                    if (tracker.currentStatusTask != null)
                                        tracker.currentStatusTask.cancel();
                                    DeleteSliverTask deleteTask = deleteSliver(tracker, state);
                                    deleteTask.addCallback((task, taskExecution, state1) -> checkNextActionAndPhase());
                                    TaskExecution taskExecution = addTaskToTaskThread(deleteTask);
                                    Platform.runLater(() -> {
                                        tracker.deleteTasks.add(new TaskWithExecution<>(deleteTask, taskExecution));
                                    });
                                }
//                                } else {
//                                    //it is already being created
//                                    log.info("ALREADY DOING nextstep: DELETE -> tracker (status=" + tracker.privateStatus + ") for " + stitchAction);
//                                }

                                break;
                            }
                            case LISTRESOURCES: {
                                synchronized (tracker) {
                                    log.info("nextstep: LISTRESOURCES -> tracker (status=" + tracker.privateStatus + ") for " + stitchAction);
                                    tracker.stitchAction = stitchAction;
                                    //dont'change tracker.setStatus(Status.);
                                    ListResourcesTask listResourcesTask = listResources(tracker, state);
                                    TaskExecution taskExecution = addTaskToTaskThread(listResourcesTask);
                                    Platform.runLater(() -> {
                                        tracker.listResourcesTasks.add(new TaskWithExecution<>(listResourcesTask, taskExecution));
                                    });
                                }
                                break;
                            }
                            default:
                                throw new RuntimeException("Unsupported StitchAction: " + stitchAction.getType());
                        }
                    }

                    noMoreActionsNeeded = !director.areAnyActionsLeft();
                    assert noMoreActionsNeeded == (director.getHopsLeft().isEmpty() || director.isGaveUp());
                }
                if (noMoreActionsNeeded) {
                    if (director.isGaveUp()) {
                        log.info("No more actions because director gave up. globalPhase="+globalPhase);
                        //no need to change state here
                    } else {
                        log.info("All hops activated. Will now switch to Provision Phase");
                        changeState(GlobalPhase.PROVISION);
                        startProvisioning();
                    }
                }

                break;
            }
            case PROVISION: {
                //start provisioning slivers if not done yet.
                startProvisioning();

                //after all is allocated, provisioning slivers
                // (might also be waiting for slivers on AMs that do not support split allocated and provision)
                synchronized (toProvision) {
                    for (StitchingActionTracker tracker : toProvision) {
                        if (tracker.privateStatus == Status.ALLOCATED) {
                            log.debug("At least one hop of  "+toProvision.size()+" not provisioned. Will keep waiting.");
                            //still waiting for a tracker to Provision
                            return;
                        }
                    }
                    log.info("All "+toProvision.size()+" hops provisioned. Will now switch to WAIT_UNTIL_READY Phase");
                }
                //all have been Provisioned
                toProvision.clear();
                changeState(GlobalPhase.WAIT_UNTIL_READY);
                //no break; //fall-through to WAIT_UNTIL_READY
            }
            case WAIT_UNTIL_READY: {
                //waiting until all slivers are ready
                synchronized (toWaitUntilReady) {
                    for (StitchingActionTracker tracker : toWaitUntilReady) {
                        if (tracker.privateStatus == Status.FAIL) {
                            log.debug("At least one hop of "+toWaitUntilReady.size()+" failed. Will giveup.");
                            giveUp("At least one hop failed");
                            return;
                        }
                        if (tracker.privateStatus != Status.READY) {
                            //still waiting for a tracker to Provision
                            log.debug("At least one hop of "+toWaitUntilReady.size()+" not ready. Will keep waiting.");
                            return;
                        }
                    }
                    log.info("All "+toWaitUntilReady.size()+" hops READY. Will now switch to SUCCESS Phase");
                }
                changeState(GlobalPhase.SUCCESS);
                toWaitUntilReady.clear();
                //no break; //fall-through to SUCCESS
            }
            case SUCCESS: {
                registerAllSliverInfo(state, slice);

                //all slivers are ready.
                countdownLatch.countDown();
                break;
            }
            case FAILED_TEMP: {
                registerAllSliverInfo(state, slice);

                toProvision.clear();
                toWaitUntilReady.clear();
                //something went wrong, and we are waiting for everything to be deleted before restarting from SCS phase
                //TODO This state is not yet used. There is a limited retry, if a sliver status becomes FAILED, the sliver is deleted and retried.
                //TODO    But there is no global retry yet. Only global giveup().
                log.info("FAILED_TEMP Phase: TODO!");
                break;
            }
            case FAILED_PERMANENT: {
                log.info("FAILED_PERMANENT Phase: gave up");
//                countdownLatch.countDown(); //done in giveup
                //something went wrong, and we gave up
                break;
            }
        }
    }

    private void createSliver(final StitchingActionTracker tracker, long delayMs, StitchingReport.State state) {
        assert tracker.privateStatus == Status.WAITING_FOR_ALLOCATE : "status is " + tracker.privateStatus;
        final AllocateSliverTask createTask = new AllocateSliverTask(tracker, state);
        createTask.addCallback(defaultTaskCallback);
        TaskExecution taskExecution;
        if (delayMs > 0)
            taskExecution = scheduleTaskToTaskThread(createTask, delayMs);
        else
            taskExecution = addTaskToTaskThread(createTask);

        Platform.runLater(() -> tracker.createTasks.add(new TaskWithExecution<>(createTask, taskExecution)));
    }

    /**
     * when calling this, you should be synchoized on tracker.
     */
    private void checkStatusUntilReady(final StitchingActionTracker tracker, StitchingReport.State state) {
        final GetStatusUntilReadyOrFailTask statusTask = new GetStatusUntilReadyOrFailTask(tracker, state);
        statusTask.addCallback(defaultTaskCallback);
        TaskExecution taskExecution = scheduleTaskToTaskThread(statusTask, TIME_BETWEEN_STATUS_CHECKS_MS);

        Platform.runLater(() -> {
            tracker.statusTasks.add(new TaskWithExecution<>(statusTask, taskExecution));
        });
    }

    private DeleteSliverTask deleteSliver(final StitchingActionTracker tracker, StitchingReport.State state) {
        return new DeleteSliverTask(tracker, state);
    }

    private ListResourcesTask listResources(final StitchingActionTracker tracker, StitchingReport.State state) {
        return new ListResourcesTask(tracker, state);
    }

    private List<AnyCredential> getUserCredentialList() {
        assert userCredentials != null;
        return userCredentials;
    }

    private List<AnyCredential> getSliceCredentialList() {
        return filterCredentials(slice.getCredentials());
    }

    private List<UserSpec> getUsers() {
//        final GeniUser user = geniUserProvider.getLoggedInGeniUser();
//
//        return Collections.singletonList(new UserSpec(user.getUserUrnString(), easyModel.getUserKeys()));
        return userSpecs;
    }

    public boolean getScsCallOk() {
        return scsCallOk.get();
    }

    public static enum Status {INIT, WAITING_FOR_DEP, WAITING_FOR_ALLOCATE, ALLOCATED, WAITING_FOR_PROVISION, PROVISIONED, FAIL, READY}

    public static interface GiveUpStitchingCallback {
        /**
         * @return true if all slivers must be deleted (= cleaning up due to failure). false if all should be left as is.
         */
        public boolean onGiveUpStitching(String reason);
    }

    public static class StitchingActionTracker {
        private final ReadOnlyListWrapper<TaskWithExecution<CreateSliverTask>> createTasks =
                new ReadOnlyListWrapper<>(FXCollections.<TaskWithExecution<CreateSliverTask>>observableArrayList());
        private final ReadOnlyListWrapper<TaskWithExecution<GetStatusUntilReadyOrFailTask>> statusTasks =
                new ReadOnlyListWrapper<>(FXCollections.<TaskWithExecution<GetStatusUntilReadyOrFailTask>>observableArrayList());
        private final ReadOnlyListWrapper<TaskWithExecution<DeleteSliverTask>> deleteTasks =
                new ReadOnlyListWrapper<>(FXCollections.<TaskWithExecution<DeleteSliverTask>>observableArrayList());
        private final ReadOnlyListWrapper<TaskWithExecution<ListResourcesTask>> listResourcesTasks =
                new ReadOnlyListWrapper<>(FXCollections.<TaskWithExecution<ListResourcesTask>>observableArrayList());
        private final ListProperty<ApiCallReply> allLoggedCalls =
                new SimpleListProperty<>(FXCollections.<ApiCallReply>observableArrayList());

        //needed to be able to cancel this
        public GetStatusUntilReadyOrFailTask currentStatusTask;

        public AbstractGeniAggregateManager.AggregateManagerReply<String> createSliverReply;
        public AbstractGeniAggregateManager.AggregateManagerReply<String> provisionReply;

        //private boolean extraSliver; //is this only needed for a link? (true if it is an extra aggregate manager)
        private ObjectProperty<Status> sharedStatus = new SimpleObjectProperty<>();
        private Status privateStatus;
        private final SfaAuthority auth;
        private int createRetries = 0;
        private long createSliverFinishMillis = 0;
        private long provisionFinishMillis = 0;
        private int statusFailureRetries = 0;
        private StitchingDirector.StitchAction stitchAction;
        private String manifestRspec;
        private boolean deletingAll = false; //when this flag is set, every non-delete call should cancel
        private Throwable failureThrowable;

        private final StitchingReport stitchingReport;
        private final StitchingReportWrappingLogger log;

        private boolean mustProvision = false;

        private StitchingActionTracker(SfaAuthority auth, StitchingReport stitchingReport, StitchingReportWrappingLogger log) {
            this.auth = auth;
            this.log = log;
            this.stitchingReport = stitchingReport;
            this.sharedStatus.set(Status.INIT);
            privateStatus = Status.INIT;
        }

        public ObjectProperty<Status> statusProperty() {
            return sharedStatus;
        }

        public SfaAuthority getAuth() {
            return auth;
        }

        public ReadOnlyListProperty<TaskWithExecution<CreateSliverTask>> createTasksProperty() {
            return createTasks.getReadOnlyProperty();
        }

        public ReadOnlyListProperty<TaskWithExecution<GetStatusUntilReadyOrFailTask>> statusTasksProperty() {
            return statusTasks.getReadOnlyProperty();
        }

        public ReadOnlyListProperty<TaskWithExecution<DeleteSliverTask>> deleteTasksProperty() {
            return deleteTasks.getReadOnlyProperty();
        }

        public Throwable getFailureThrowable() {
            return failureThrowable;
        }

        public Status getStatus() {
            return sharedStatus.get();
        }

        private void setStatus(final Status s) {
            if (privateStatus != s) {
                log.debug("Status changed to " + s + " on " + auth.getUrnString());
                stitchingReport.getCurState().info("SliverReadyTracker status changed to " + s + " on " + auth.getUrnString());
            }
            privateStatus = s;
            if (Platform.isFxApplicationThread())
                sharedStatus.set(s);
            else
                Platform.runLater(() -> sharedStatus.set(s));
        }

        public StatusDetails getLatestSliverStatusDetails() {
            if (statusTasks.isEmpty())
                return null;

            return statusTasks.get(statusTasks.size() - 1).getTask().getStatusDetails();
        }

        public Collection<StatusDetails> getAllSliverStatusDetails() {
            return statusTasks.stream()
                    .map(twe -> twe.getTask().getStatusDetails())
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
        }
    }

    public class SCSRequestTask extends Task {
        private final SfaAuthority stitcherAuth;
        private final GeniUser user = geniUserProvider.getLoggedInGeniUser();
        private final StitchingReport.State state;

        public SCSRequestTask(StitchingReport.State state) {
            super("ParallelStitcher StitchingComputationService call");
            this.state = state;

            this.stitcherAuth = getStitchingAuthority();
            assert stitcherAuth != null;
        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            log.debug("Sedning StitchingComputationService request");
            StitchingComputationService scs =
                    new StitchingComputationService(JavaFXLogger.wrap(easyModel.getLogger(), singleTask, state), jFedPreferences);

            assert user != null;
            assert slice != null;
            assert stitcherAuth != null;

            //see FEDIBBTDEV-1334: ION and others see "expires" in the request RSpec as the requested sliver expiration
            //so if not yet specified, we add it here
            String modifiedRequestRspec = null;
            if (requestedSliverExpirationDate != null && stitcherAuth.hasSpecialCase(SfaAuthority.SpecialCase.ADD_EXPIRES_ATTRIBUTE_TO_REQUEST)) {
                ModelRspec modelRspec = originalRspecRequest.getModelRspec();
//                if (modelRspec.getExpiresString() == null) { //FEDIBBTDEV-1410 always overwrite
                modelRspec = modelRspec.copy();
                modelRspec.setExpires(Date.from(requestedSliverExpirationDate));
                modifiedRequestRspec = modelRspec.toGeni3Rspec();
//                }
            }

            try {
                SfaConnection con =
                        (SfaConnection) connectionProvider.getConnectionByAuthority(user, stitcherAuth, StitchingComputationService.class);
                StitchingComputationService.SCSReply<StitchingComputationService.ComputePathResult> res =
                        scs.computePath(con, slice.getUrnString(),
                                modifiedRequestRspec == null ? originalRspecRequest.getRspecXmlString()
                                        : modifiedRequestRspec,
                                null);

                if (res != null) {
                    if (res.getValue() != null) {
                        scsResult.set(res.getValue());
                        log.info("SCS call success");
                        onScsTaskSuccess();
                    } else if (res.getOutput() != null) {
                        throw new JFedHighLevelException(String.format("SCS returned a failure: %s", res.getOutput()));
                    } else {
                        throw new JFedHighLevelException("Could not process SCS response: response without value or output");
                    }
                } else {
                    throw new JFedHighLevelException("Could not process SCS response: is NULL");
                }
            } catch (JFedException e) {
                onScsTaskFail();
                log.error("SCS call failed", e);
                giveUp("SCS call failed");
                throw e;
            }
        }

        /**
         * This method processes the result of the StitchingComputationService.
         * It initializes the next steps by creating SliverReadyTrackers.
         * After that activateReadyHops is called to determine which slivers are
         * initialized in a first step.
         */
        private void onScsTaskSuccess() {
            StitchingComputationService.ComputePathResult res = scsResult.get();
            assert res != null;

            synchronized (director) {
                director.setComputePathResult(res);

                //also update the slice request rspec with the enriched request-rspec from the SCS
                slice.setRequestRspec(new RequestRspecSource(new StringRspec(res.getServiceRspec()), ModelRspecType.FX));

                Platform.runLater(() -> {
                    scsCallOk.set(true);
                    scsCallFinished.set(true);
                });

                for (SfaAuthority auth : director.getAllAuthorities()) {
                    assert auth != null;
                    log.debug("Creating SliverReadyTracker for " + auth);
                    final StitchingActionTracker tracker = new StitchingActionTracker(auth, stitchingReport, log);
                    trackers.put(auth, tracker);

                    tracker.setStatus(Status.WAITING_FOR_DEP);
                    Platform.runLater(() -> sliverTrackers.add(tracker));
                }
            }
            changeState(GlobalPhase.ALLOCATE);
            checkNextActionAndPhase();
        }

        private void onScsTaskFail() {
            Platform.runLater(() -> {
                stitchingFailed.set(true);
                scsCallFinished.set(true);
                scsCallOk.set(false);
            });

            countdownLatch.countDown();
        }
    }

    public abstract class CreateSliverTask extends Task {
        protected final StitchingActionTracker tracker;
        protected final GeniUser user = geniUserProvider.getLoggedInGeniUser();

        private CreateSliverTask(String name, final StitchingActionTracker tracker) {
            super(name);

            this.tracker = tracker;
            assert tracker != null;
            assert tracker.stitchAction != null;
        }

        public SfaAuthority getAuthority() {
            return tracker.getAuth();
        }
    }


    public class AllocateSliverTask extends CreateSliverTask {
        private final StitchingReport.State state;
        private boolean supportsAllocate;

        private AllocateSliverTask(final StitchingActionTracker tracker, StitchingReport.State state) {
            super("ParallelStitcher Allocate or CreateSliver call", tracker);
            this.state = state;
            assert state != null;

            assert state.getName().equals("ALLOCATE");
            log.debug("adding Allocate or CreateSliver task for " + tracker.getAuth().getUrnString());
        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            assert state.getName().equals("ALLOCATE");
            try {
                if (tracker.deletingAll) return;

                //TODO make CreateSliver-call used here more generic

                log.info("Calling CreateSliver on " + tracker.stitchAction.getAuthority().getUrn());
                assert tracker.privateStatus == Status.WAITING_FOR_ALLOCATE :
                        "tracker status is " + tracker.privateStatus;
//                tracker.setStatus(Status.WAITING_FOR_ALLOCATE);

                assert !tracker.auth.isFake();
                assert tracker.auth.isReal();

                String manifestRspec;
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, tracker.auth, state);
                supportsAllocate = amWrapper.hasSeperateAllocateAndProvision();

                Instant usedExpirationDate =
                        amWrapper.isExirationdateSetByCreateSliver() ? requestedSliverExpirationDate
                                : null; //don't set date here if CreateSliver doesn't set it itself
                if (tracker.auth.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_RENEW_AFTER_CREATESLIVER))
                    usedExpirationDate = requestedSliverExpirationDate;

                AbstractGeniAggregateManager.AggregateManagerReply reply;
                try {
                    String requestRspec = tracker.stitchAction.getRequestRspec();


                    assert requestRspec != null;
                    log.debug("Request RSpec stitching info for " + tracker.auth.getUrnString() + ":\n" + HtmlLogOutput.makeStitchingOverview(requestRspec));
                    if (supportsAllocate) {
                        tracker.mustProvision = true;
                        AggregateManagerWrapper.AllocateReply allocateReply = amWrapper.allocate(slice.getUrn(), getSliceCredentialList(), requestRspec,
                                usedExpirationDate == null ? null
                                        : Date.from(usedExpirationDate)/*expirationDate*/);
                        manifestRspec = allocateReply == null ? null : allocateReply.getManifest();
                        reply = amWrapper.getLastReply();
                        assert reply != null;
                    } else {
                        tracker.mustProvision = false;
                        manifestRspec =
                                amWrapper.createSliver(slice.getUrn(), getSliceCredentialList(), requestRspec, getUsers(),
                                        usedExpirationDate == null ? null
                                                : Date.from(usedExpirationDate)/*expirationDate*/);
                        reply = amWrapper.getLastReply();
                        assert reply != null;
                    }
                } catch (AggregateManagerWrapper.VlanUnavailableException e) {
                    log.info("vlan was not available", e);
                    manifestRspec = null;
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                    assert StitchingDirector.isUnavailableVlan(reply);
                } catch (AggregateManagerWrapper.AnyVlanUnsupportedException e) {
                    log.info("any was not supported", e);
                    manifestRspec = null;
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                    assert StitchingDirector.isAnyNotSupported(reply);
                }

                final AbstractGeniAggregateManager.AggregateManagerReply reply_final = reply;
                Platform.runLater(() -> tracker.allLoggedCalls.add(reply_final));

                tracker.createSliverReply = reply;
//                    if (!reply.getXmlRpcGeniResponseCode().isSuccess())
//                        throw new RuntimeException("Call not succesfull: " + reply.getXmlRpcGeniResponseCode());
                if (reply != null && reply.getGeniResponseCode().isSuccess()) {
                    tracker.manifestRspec = manifestRspec;
                }

                processAllocateResult();
            } catch (JFedException e) {
                assert tracker != null;
                log.error("Failed to create sliver at " + tracker.auth.getName(), e);

                tracker.failureThrowable = e;

                if (StitchingDirector.isUnrecoverableError(e) == Boolean.TRUE) {
                    log.error("Reply was recognized as unrecoverable error. This failure is permanent -> giving up stitching.", e);
                    doGiveUp();
                } else
                    doRetry();
                throw e;
            }
        }

        private void doRetry() {
            log.trace("doRetry()");

            if (tracker.deletingAll) return;

            if (tracker.createRetries++ < MAX_CREATE_RETRIES) {
                tracker.setStatus(Status.FAIL);

                log.info("ALLOCATE.doRetry: CreateSliver failed (createRetries=" + tracker.createRetries + "), " +
                        "will delete and try again in " + TIME_BEFORE_CREATESLIVER_RETRY_MS + "ms " +
                        "now=" + HtmlLogOutput.dateToString(new Date()));

                StitchingDirector.AllocateVerdict verdict =
                        director.reportMustDelete(tracker.auth, false/*vlansUnavailable*/, true/*forceAllocateFinished*/);
                switch (verdict) {
                    case GIVEUP: {
                        log.info("ALLOCATE.doRetry: director gave up");
                        giveUp("Director gave up when we requested retry: ");
                        break;
                    }
                    case DIRECTOR_ACTION: {
                        //tracker.setStatus(Status.); //keep status
                        log.info("ALLOCATE.doRetry: director wants to try again");
                        tracker.privateStatus = Status.FAIL;

                        //change back to ALLOCATE GlobalPhase if needed
                        if (globalPhase != GlobalPhase.ALLOCATE && globalPhase != GlobalPhase.FAILED_PERMANENT && globalPhase != GlobalPhase.FAILED_TEMP) {
                            globalPhase = GlobalPhase.ALLOCATE;
                            if (!stitchingReport.getCurState().getName().equals("ALLOCATE")) {
                                changeState(GlobalPhase.ALLOCATE);
                            }
                            log.debug("Changed state back to ALLOCATE");
                        }

                        break;
                    }
                    default:
                        throw new RuntimeException("implementation error: director.reportMustDelete should never return " + verdict);
                }
                checkNextActionAndPhase();
            } else {
                doGiveUp();
            }
        }

        private void doGiveUp() {
            log.info("CreateSliver failed, will delete and not try again");
            tracker.setStatus(Status.FAIL);
            giveUp("CreateSliver failed on " + tracker.getAuth() + " (permanent failure detected.)");
        }

        private void processAllocateResult() {
            log.trace("processAllocateResult()");

            if (tracker.deletingAll) {
                log.warn("CreateSliver result NOT passed to director because already deleting slice: code=" + tracker.createSliverReply.getCode() + " output=" + tracker.createSliverReply.getOutput());
                return;
            }

            assert tracker.stitchAction != null;
            assert tracker.createSliverReply != null;

            //try to handle bug gracefully
            if (tracker.stitchAction == null || tracker.createSliverReply == null) {
                log.error("BUG! Internal error in onCreateSliverTaskSuccess() " +
                        "tracker.stitchAction=" + (tracker.stitchAction == null ? "null"
                        : "tracker for " + tracker.stitchAction.getAuthority().getUrnString()) + " " + " " +
                        "tracker.reply=" + (tracker.createSliverReply == null ? "null" : "non-null"));
                doRetry();
                return;
            }

            StitchingDirector.AllocateVerdict verdict;
            synchronized (director) {
                assert state.getName().equals("ALLOCATE");
                verdict = director.processAllocateResult(tracker.stitchAction, tracker.createSliverReply);
            }
            log.info("CreateSliver result passed to director: code=" + tracker.createSliverReply.getCode() +
                    " output=" + tracker.createSliverReply.getOutput() + " -> director verdict=" + verdict);

            switch (verdict) {
                case GIVEUP: {
                    doGiveUp();
                    break;
                }
                case CALLER_DELAYED_RETRY: {
                    doRetry();
                    break;
                }
                case DIRECTOR_ACTION: {
                    //most likely case: must try again with different VLAN.
                    tracker.setStatus(Status.INIT);
                    checkNextActionAndPhase();
                    break;
                }
                case OK: {
                    log.debug("Overview of hop data after CreateSliver call: \n" + director.getSuggestedAvailableOverviewsString());

                    //TODO: I think we can wait with all this and do it when the next phase starts. Reason: the director might delete this again later!
                    synchronized (tracker) {
                        if (tracker.mustProvision) {
                            assert supportsAllocate;
                            //only allocate was success. Need to provision later
                            tracker.setStatus(Status.ALLOCATED);

                            toProvision.add(tracker);
                            toWaitUntilReady.add(tracker); //do this here already, to avoid any possible race condition in checkNextActionAndPhase();
                        } else {
                            assert !supportsAllocate;
                            toWaitUntilReady.add(tracker); //do this first to wavoid race condition in checkNextActionAndPhase();

                            tracker.setStatus(Status.PROVISIONED);
                            tracker.createSliverFinishMillis = System.currentTimeMillis();

                            checkStatusUntilReady(tracker, state);
                        }
                    }
                    checkNextActionAndPhase();
                    break;
                }
                default:
                    throw new RuntimeException("BUG! unhandled verdict " + verdict);
            }
        }

        public boolean isSupportsAllocate() {
            return supportsAllocate;
        }
    }

    public class GetStatusUntilReadyOrFailTask extends Task {
        private final StitchingReport.State state;

        private static final int MAX_FAIL_COUNT = 5;
        private final StitchingActionTracker tracker;
        private int failCount = 0;
        private int triesCount = 0;
        private String status;
        private StatusDetails statusDetails;
        private AbstractGeniAggregateManager.AggregateManagerReply amReply;
        private boolean finished = false;

        private boolean cancelled = false;

        /**
         * the caller of this should be synchronized on tracker
         */
        public void cancel() {
            cancelled = true;
            finished = true;
            tracker.currentStatusTask = null;
        }

        public GetStatusUntilReadyOrFailTask(final StitchingActionTracker tracker, StitchingReport.State state) {
            super("ParallelStitcher CheckStatus call");
            this.tracker = tracker;
            this.state = state;
            tracker.currentStatusTask = this;

            log.debug("adding CheckStatus task for " + tracker.getAuth().getUrnString());
        }


        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            try {
                synchronized (tracker) {
                    if (cancelled) {
                        log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                        return;
                    }
                    //check if we still should continue
                    if (tracker.deletingAll) {
                        assert !finished;
                        finished = true;
                        return;
                    }
                }

                triesCount++;
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, tracker.auth, state);

                boolean busy = false;

                try {
                    statusDetails = amWrapper.status(slice.getUrn(), getSliceCredentialList());
                    amReply = amWrapper.getLastReply();
                } catch (JFedException ex) {
                    if (cancelled) {
                        log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                        return;
                    }
                    if (ex.getGeniResponseCode() != null && ex.getGeniResponseCode().isBusy()) {
                        log.warn("Got an busy-error while fetching status of sliver. Skipping...");
                        busy = true;
                    } else {
                        throw ex;
                    }
                }
                if (cancelled) {
                    log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                    return;
                }

                if (!busy && statusDetails == null) {
                    failCount++;

                    if (failCount < MAX_FAIL_COUNT) {
                        log.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                                " Trying again in " + TIME_BETWEEN_STATUS_CHECKS_MS + " ms...");
                        scheduleTaskToTaskThread(this, TIME_BETWEEN_STATUS_CHECKS_MS);
                    } else {
                        log.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                                " Will stop checking status...");
                        this.status = "Fetching status failed. Aborting checks!";

                        onSliverStatusCallFailed(null);

                        finished = true;
                    }
                    return;
                }

                if (busy || statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNKNOWN) || statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.CHANGING)) {
                    //check if timeout has passed
                    long now = System.currentTimeMillis();
                    long diff = now - (tracker.mustProvision ? tracker.provisionFinishMillis
                            : tracker.createSliverFinishMillis);
                    if (diff < STATUS_TIMEOUT_MS) {
                        log.info("Status of Slice " + slice.getUrnString() + " is \"" + statusDetails + "\": will check again in " + TIME_BETWEEN_STATUS_CHECKS_MS + " ms.");
                        scheduleTaskToTaskThread(this, TIME_BETWEEN_STATUS_CHECKS_MS);
                    } else {
                        //timeout has expired. Abort!
                        log.warn("Slice " + slice.getUrnString() + " has exceeded it's maximum setup time of " + STATUS_TIMEOUT_MS + " ms. (total time taken was " + diff + ") Aborting...");
                        this.status = "Maximum setup time has been exceeded. Aborting...";
                        giveUp("Maximum setup time (" + STATUS_TIMEOUT_MS + " ms) has been exceeded while waiting for sliver(s) at " + tracker.getAuth() + " to become ready.");
                        finished = true;

                    }
                } else if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.READY)) {
                    synchronized (tracker) {
                        if (cancelled) {
                            log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                            return;
                        }
                        tracker.setStatus(Status.READY);

                        addTaskToTaskThread(highLevelController.describe(tracker.auth, getSlice()));
                    }

                    log.debug("tracker status became READY on " + tracker.auth.getUrnString());
                    checkNextActionAndPhase();

                    if (!amWrapper.isExirationdateSetByCreateSliver() && requestedSliverExpirationDate != null) {
                        boolean onlyIfNeeded = true;

                        if (tracker.auth.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_RENEW_AFTER_CREATESLIVER))
                            onlyIfNeeded = false; //always call renew

                        if (cancelled) {
                            log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                            return;
                        }

                        HighLevelController.RenewSliversTask renewSliversTask =
                                highLevelController.renewSliversAtAuthority(
                                        getSlice(), requestedSliverExpirationDate,
                                        tracker.auth, null/*slivers (null = all)*/,
                                        onlyIfNeeded/*only if renew is needed*/);
                        addTaskToTaskThread(renewSliversTask);

                        if (cancelled) {
                            log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                            return;
                        }
                    }

                    finished = true;

                } else if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL)) {
                    boolean permanentFailure = amReply == null ? true : StitchingDirector.isUnrecoverableError(amReply);
                    if (permanentFailure)
                        onPermanentSliverFailedStatus();
                    else
                        onSliverFailedStatus();

                    finished = true;
                }
            } catch (JFedException ex) {
                onSliverStatusCallFailed(ex);
                finished = true;

                throw ex;
            }
        }

        public StatusDetails getStatusDetails() {
            return statusDetails;
        }

        public long getMsUntilNextCheck() {
            if (!futureTasks.isEmpty()) {
                return futureTasks.get(0).getTimeLeftMs();
            } else
                return 0;

        }

        private void onSliverStatusCallFailed(@Nullable JFedException exception) {
            synchronized (tracker) {
                if (cancelled) {
                    log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                    return;
                }

                //ex might be null, if there was an unexpected failure that did not have to do with an exception
                log.error("Got an error while checking status of slice " + slice.getUrnString(), exception);
                boolean permanentFailure = exception == null ?
                        false : //not sure if permanent or not, so retry is ok
                        StitchingDirector.isUnrecoverableError(exception) == Boolean.TRUE;

                //handle status call failure (not sliver)
                if (!permanentFailure || tracker.statusFailureRetries++ < MAX_STATUS_RETRIES) {
                    log.warn("Checking of status of Slice " + slice.getUrnString() + " FAILED: will check again in " + TIME_BETWEEN_STATUS_CHECKS_MS + " ms. status=" + status);
                    scheduleTaskToTaskThread(this, TIME_BETWEEN_STATUS_CHECKS_MS);
                } else {
                    giveUp("Status check failed for sliver(s) on " + tracker.getAuth() + " (permanent failure detected.)");
                    finished = true;
                }
            }

            checkNextActionAndPhase();
        }


        /**
         * status call successful, but sliver status is failure
         */
        private void onSliverFailedStatus() {
            status = "The provisioning of the resources failed";

            synchronized (tracker) {
                if (cancelled) {
                    log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                    return;
                }

                if (tracker.createRetries++ < MAX_CREATE_RETRIES) {
                    //delete and try again.
                    log.debug("STATUS is FAILED");

                    boolean vlansUnavailable = amReply == null ? false : StitchingDirector.isUnavailableVlan(amReply);
                    boolean permanentFailure = amReply == null ? true : StitchingDirector.isUnrecoverableError(amReply);
                    if (permanentFailure || !vlansUnavailable) {
                        if (permanentFailure)
                            log.info("STATUS is FAILED: detected permanent failure, giving up.");
                        else {
                            if (!vlansUnavailable)
                                log.info("STATUS is FAILED: generic failure, giving up.");
                        }
                        giveUp("Received sliver Status had unrecoverable error: " + statusDetails);
                    } else {
                        log.debug("STATUS is FAILED: asking director to delete sliver");
                        StitchingDirector.AllocateVerdict verdict = director.reportMustDelete(tracker.getAuth(), vlansUnavailable, false/*forceAllocateFinished*/);
                        log.debug("STATUS is FAILED: director verdict = "+verdict);
                        switch (verdict) {
                            case GIVEUP: {
                                log.info("STATUS is FAILED: director gave up");
                                giveUp("Director gave up because of failed sliver Status: "+statusDetails);
                                break;
                            }
                            case DIRECTOR_ACTION: {
                                //tracker.setStatus(Status.); //keep status
                                log.info("STATUS is FAILED: director wants to try again");

                                //change back to ALLOCATE GlobalPhase if needed
                                if (globalPhase != GlobalPhase.ALLOCATE && globalPhase != GlobalPhase.FAILED_PERMANENT && globalPhase != GlobalPhase.FAILED_TEMP) {
                                    globalPhase = GlobalPhase.ALLOCATE;
                                    if (!stitchingReport.getCurState().getName().equals("ALLOCATE")) {
                                        changeState(GlobalPhase.ALLOCATE);
                                    }
                                    log.debug("Changed state back to ALLOCATE");
                                }

                                break;
                            }
                            default:
                                throw new RuntimeException("implementation error: director.reportMustDelete should never return the verdict " + verdict);
                        }
                    }
                } else {
                    String attempts = MAX_CREATE_RETRIES == 0 ? ""
                            : " (" + tracker.createRetries + " attempts to delete sliver and try create again have already been unsuccessful)";
                    log.error("Sliver status is FAILED. Deleting and not trying again. " + attempts);
                    tracker.setStatus(Status.FAIL);
                    giveUp("Sliver became failed on " + tracker.getAuth() + attempts);
                }
            }

            checkNextActionAndPhase();
        }

        /**
         * status call successful, but sliver status is permanent failure
         */
        private void onPermanentSliverFailedStatus() {
            status = "The provisioning of the resources failed permanently";

            synchronized (tracker) {
                if (cancelled) {
                    log.debug("GetStatusUntilReadyOrFailTask saw cancelled: stopping right now");
                    return;
                }

                log.info("Sliver status is FAILED. Deleting and not trying again.");
                tracker.setStatus(Status.FAIL);
                giveUp("Sliver became failed on " + tracker.getAuth() + " (permanent failure detected)");
            }

            checkNextActionAndPhase();
        }

        public SfaAuthority getAuthority() {
            return tracker.getAuth();
        }

        public boolean isFinished() {
            return finished;
        }

        public int getTriesCount() {
            return triesCount;
        }

        public String getStatus() {
            return status;
        }
    }

    public class DeleteSliverTask extends Task {
        private final StitchingActionTracker tracker;
        private final GeniUser user = geniUserProvider.getLoggedInGeniUser();
        private final StitchingReport.State state;

        public DeleteSliverTask(final StitchingActionTracker tracker, StitchingReport.State state) {
            super("ParallelStitcher DeleteSliver call");
            this.tracker = tracker;
            this.state = state;

            log.debug("making DeleteSliver task for " + tracker.getAuth().getUrnString());

        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            try {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, tracker.auth, state);
                amWrapper.deleteSliver(slice.getUrn(), getSliceCredentialList());
            } catch (JFedException e) {
                log.error("Failed to delete sliver at " + tracker.auth.getName(), e);
                throw e;
            } finally {
                //let the director know of this Delete. (if it failed, a retry will be needed anyway)
                director.processDeleteResult(tracker.auth);
                tracker.setStatus(Status.INIT);
            }

            try {
                unregisterSliverInfo(slice, tracker.auth);
            } catch (JFedException e) {
                log.error("Failed to delete SLIVER_INFO (error ignored)", e);
                //ignore
            }
        }

        public SfaAuthority getAuthority() {
            return tracker.getAuth();
        }
    }

    public class ListResourcesTask extends Task {
        private final StitchingActionTracker tracker;
        private final GeniUser user = geniUserProvider.getLoggedInGeniUser();
        private final StitchingReport.State state;

        public ListResourcesTask(final StitchingActionTracker tracker, StitchingReport.State state) {
            super("ParallelStitcher ListResourcesTask call");
            this.tracker = tracker;
            this.state = state;

            log.debug("making ListResources task for " + tracker.getAuth().getUrnString());
        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            String advertisementRspec = null;
            try {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, tracker.auth, state);
                List<AnyCredential> userCredential = getUserCredentialList();
                boolean available = true;
                log.debug("calling ListResources on " + tracker.getAuth().getUrnString());
                advertisementRspec = amWrapper.listResources(userCredential, available);
            } catch (Exception e) {
                log.error("Failed ListResources at " + tracker.auth.getName(), e);
                throw e;
            } finally {
                //let the director know of this ListResources (if it failed, a null advertisementRspec is reported)
                director.processListResourcesResult(tracker.auth, advertisementRspec);

                checkNextActionAndPhase();
            }
        }

        public SfaAuthority getAuthority() {
            return tracker.getAuth();
        }
    }


    private void registerAllSliverInfo(StitchingReport.State state, Slice slice) {
        log.debug("registerAllSliverInfo for "+director.getAllAuthorities().size()+" auths: "+director.getAllAuthorities());
        for (SfaAuthority auth : director.getAllAuthorities()) {
            registerSliverInfo(state, slice, auth, slice.getManifestRspec());
        }
    }

    private void registerSliverInfo(StitchingReport.State state, Slice slice, SfaAuthority auth, ManifestRspecSource manifestRspec) {
        assert slice != null;
        assert auth != null;

        try {
            Set<Sliver> createdSlivers = new HashSet();
            Collection<Sliver> manifestSlivers =
                    manifestRspec == null ? null :
                            EasyModelAbstractListener.findSliversInManifest(easyModel, auth, slice.getUrnString(), manifestRspec);
            //add slivers found by easyModel due to SliverStatus and Status calls. //AL2S is one case where this is need, see FEDIBBTDEV-1392
            Set<Sliver> statusSlivers =
                    slice.getSlivers() == null ? Collections.emptySet() :
                            slice.getSlivers().stream().filter(s -> {
                                        if (s.getAuthority() == null || auth == null)
                                            return s.getAuthority() == auth;
                                        else
                                            return s.getAuthority().equals(auth);
                                    }
                            )/*. map(Sliver::getUrn).filter(Objects::nonNull)*/.collect(Collectors.toSet());

            if (manifestSlivers != null) createdSlivers.addAll(manifestSlivers);
            if (statusSlivers != null) createdSlivers.addAll(statusSlivers);

            if (!createdSlivers.isEmpty()) {
                log.debug("Registering "+createdSlivers.size()+" known slivers for "+auth.getUrnString());
                if (manifestSlivers != null)
                    log.debug("   "+manifestSlivers.size()+" slivers from manifest: "+manifestSlivers.stream().map(Sliver::getUrn).collect(Collectors.toList()));
                else
                    log.debug("    no slivers from manifest.");
                if (statusSlivers != null)
                    log.debug("   "+statusSlivers.size()+  " slivers from EasyModel: "+statusSlivers.stream().map(Sliver::getUrn).collect(Collectors.toList()));
                else
                    log.debug("    no slivers from EasyModel.");
                TaskExecution t =
                        highLevelController.registerSliversAtSATaskExecution(createdSlivers,
                                null/*extra task to depend on (not needed: create sliver has returned already)*/,
                                state/*extra result listener*/);
            }
        } catch (Exception e) {
            log.error("Failed to register SLIVER_INFO (error ignored)", e);
        }
    }

    private void unregisterSliverInfo(Slice slice, SfaAuthority auth) throws JFedException {
        assert slice != null;
        assert auth != null;

        //TODO: would perhaps be good if delete task depended on any matching register task.
        TaskExecution t =
                highLevelController.unregisterSliversAtSATaskExecution(slice, auth, null/*extra task to depend on (not needed: delete sliver has returned already)*/);
    }


    private void startProvisioning() {
        assert globalPhase == GlobalPhase.PROVISION;

        StitchingReport.State provisionState = stitchingReport.getCurState();

        synchronized (toProvision) {
            log.debug("Starting Provision on "+toProvision.size()+" authorities");
            for (StitchingActionTracker tracker : toProvision) {
                if (tracker.privateStatus == Status.ALLOCATED) {
                    assert tracker.mustProvision;
                    tracker.setStatus(Status.WAITING_FOR_PROVISION);
                    final ProvisionSliverTask provisionTask = new ProvisionSliverTask(tracker, provisionState);
                    provisionTask.addCallback(defaultTaskCallback);
                    TaskExecution taskExecution = addTaskToTaskThread(provisionTask);

                    Platform.runLater(() -> tracker.createTasks.add(new TaskWithExecution<>(provisionTask, taskExecution)));
                }
            }
        }
    }

    public class ProvisionSliverTask extends CreateSliverTask {
        private final StitchingReport.State state;

        private ProvisionSliverTask(final StitchingActionTracker tracker, StitchingReport.State state) {
            super("ParallelStitcher Provision call", tracker);
            this.state = state;

            log.debug("adding Provision task for " + tracker.getAuth().getUrnString());
        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            try {
                if (tracker.deletingAll) return;

                //TODO make Provision-call used here more generic

                log.info("Calling Provision on " + tracker.stitchAction.getAuthority().getUrn());
                assert tracker.privateStatus == Status.WAITING_FOR_PROVISION :
                        "Status is not WAITING_FOR_PROVISION but " + tracker.privateStatus;
                assert tracker.mustProvision;
                //tracker.setStatus(Status.CHANGING);

                String manifestRspec;
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, tracker.auth, state);

                Date usedExpirationDate =
                        amWrapper.isExirationdateSetByCreateSliver() ? (requestedSliverExpirationDate == null ? null
                                : Date.from(requestedSliverExpirationDate))
                                : null; //don't set date here if CreateSliver doesn't set it itself
                if (tracker.auth.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_RENEW_AFTER_CREATESLIVER))
                    usedExpirationDate =
                            requestedSliverExpirationDate == null ? null : Date.from(requestedSliverExpirationDate);

                AbstractGeniAggregateManager.AggregateManagerReply reply;
//                try {
                assert amWrapper.hasSeperateAllocateAndProvision();
                assert tracker.mustProvision;
                manifestRspec =
                        amWrapper.provisionAndStart(slice.getUrn(), getSliceCredentialList(), getUsers(), usedExpirationDate/*expirationDate*/);
                reply = amWrapper.getLastReply();
//                }
                //these should not occur on Provision. And if they do, we'll treat them as regular exceptions.
//                catch (AggregateManagerWrapper.VlanUnavailableException e) {
//                    log.info("vlan was not available", e);
//                    manifestRspec = null;
//                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
//                } catch (AggregateManagerWrapper.AnyVlanUnsupportedException e) {
//                    log.info("any was not supported", e);
//                    manifestRspec = null;
//                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
//                }

                final AbstractGeniAggregateManager.AggregateManagerReply reply_final = reply;
                Platform.runLater(() -> tracker.allLoggedCalls.add(reply_final));

                tracker.provisionReply = reply;
                if (reply != null && reply.getGeniResponseCode().isSuccess()) {
                    tracker.manifestRspec = manifestRspec;
                    onProvisionTaskSuccess();
                } else
                    onProvisionTaskFailed(null, reply);

            } catch (JFedException e) {
                log.error("Failed to Provision sliver at " + tracker.auth.getName(), e);

                onProvisionTaskFailed(e, null);
                throw e;
            }
        }

        private void onProvisionTaskFailed(Throwable throwable, AbstractGeniAggregateManager.AggregateManagerReply reply) {
            log.trace("onProvisionTaskFailed(t, r)");

            if (tracker.deletingAll) return;

            tracker.failureThrowable = throwable;

            log.info("CreateSliver failed, will delete and not try again");
            tracker.setStatus(Status.FAIL);
            giveUp("CreateSliver failed on " + tracker.getAuth() + " (permanent failure detected.)");
        }

        private void onProvisionTaskSuccess() {
            log.trace("onProvisionTaskSuccess()");

            if (tracker.deletingAll) {
                log.warn("CreateSliver result NOT passed to director because already deleting slice:" +
                        " code=" + tracker.provisionReply.getCode() +
                        " output=" + tracker.provisionReply.getOutput());
                return;
            }

            /* note: while the task was successful, the call might be an error. The stitching director decides about that. */
            assert tracker.stitchAction != null;
            assert tracker.provisionReply != null;

            //handle bug
            if (tracker.stitchAction == null || tracker.provisionReply == null) {
                log.error("Internal error in onProvisionTaskSuccess() " +
                        "tracker.stitchAction=" + (tracker.stitchAction == null ? "null"
                        : "tracker for " + tracker.stitchAction.getAuthority().getUrnString()) + " " + " " +
                        "tracker.reply=" + (tracker.provisionReply == null ? "null" : "non-null"));
                onProvisionTaskFailed(null, null);
                return;
            }

            assert tracker.provisionReply.getGeniResponseCode().isSuccess();

            synchronized (tracker) {
                tracker.provisionFinishMillis = System.currentTimeMillis();

                tracker.setStatus(Status.PROVISIONED);
                checkStatusUntilReady(tracker, state);
            }

            checkNextActionAndPhase();
        }
    }


    /**
     * @param timeout
     * @param unit
     * @return {@code true} if the count reached zero and {@code false}
     * if the waiting time elapsed before the count reached zero
     * @throws InterruptedException
     */
    public boolean waitUntilDone(long timeout, TimeUnit unit) throws InterruptedException {
        long startMs = System.currentTimeMillis();
        log.debug("waitUntilDone -> start at " + HtmlLogOutput.dateToString(startMs));
        long maxWait = TimeUnit.MILLISECONDS.convert(timeout, unit);
        boolean res = countdownLatch.await(timeout, unit);
        if (res == false) {
            log.debug("waitUntilDone -> timeout waiting for countdownLatch at " + HtmlLogOutput.dateToString(new Date()));
            return res;
        }
        //wait for all tasks to finish
        while (true)
            synchronized (taskCount) {
                //all tasks ready?
                if (taskCount.get() <= 0) {
                    log.debug("waitUntilDone -> all ready at " + HtmlLogOutput.dateToString(new Date()));
                    return true;
                }

                //if not, wait until update, unless timeout reached
                long now = System.currentTimeMillis();
                long waitedMs = now - startMs;
                if (waitedMs < maxWait) {
                    long left = maxWait - waitedMs;
                    taskCount.wait(left);
                } else {
                    log.debug("waitUntilDone -> timeout waiting for " + taskCount.get() + " tasks at " + HtmlLogOutput.dateToString(now));
                    return false;//timeout
                }
            }
    }


    public GlobalPhase getGlobalPhase() {
        return globalPhase;
    }

    public StitchingReport getStitchingReport() {
        return stitchingReport;
    }


//    public void trace(StitchingReport.State state, String message) {
//        log.trace(message);
//        state.debug(message);
//    }
//
//    public void debug(StitchingReport.State state, String message) {
//        log.debug(message);
//        state.debug(message);
//    }
//
//    public void info(StitchingReport.State state, String message) {
//        log.info(message);
//        state.info(message);
//    }
//
//    public void warn(StitchingReport.State state, String message) {
//        log.warn(message);
//        state.warn(message);
//    }
//
//    public void error(StitchingReport.State state, String message) {
//        log.error(message);
//        state.error(message);
//    }
//
//    public void info(StitchingReport.State state, String message, Throwable t) {
//        log.info(message, t);
//        state.info(message, t);
//    }
//
//    public void warn(StitchingReport.State state, String message, Throwable t) {
//        log.warn(message, t);
//        state.warn(message, t);
//    }
//
//    public void error(StitchingReport.State state, String message, Throwable t) {
//        log.error(message, t);
//        state.error(message, t);
//    }

    /**
     * Change global state, and report state. Etc.
     *
     * @returns new report state
     */
    private StitchingReport.State changeState(GlobalPhase newPhase) {
        globalPhase = newPhase;
        StitchingReport.State reportState = stitchingReport.changeState(globalPhase + "");
        log.setState(reportState);
        stitchingDirectorLog.setState(reportState);
        return reportState;
    }


    private AtomicInteger taskCount = new AtomicInteger(0);

    private void registerTaskDone(Task t) {
        synchronized (taskCount) {
            int i = taskCount.decrementAndGet();
            if (i == 0)
                taskCount.notify();
        }
    }

    private TaskExecution scheduleTaskToTaskThread(Task t, long delayMs) {
        synchronized (taskCount) {
            t.addCallback((task, taskExecution, state) -> registerTaskDone(t));
            TaskExecution taskExecution = taskThread.scheduleTask(t, delayMs);
            taskCount.incrementAndGet();

            return taskExecution;
        }
    }

    private TaskExecution addTaskToTaskThread(Task t) {
        synchronized (taskCount) {
            t.addCallback((task, taskExecution, state) -> registerTaskDone(t));
            TaskExecution taskExecution = taskThread.addTask(t);
            taskCount.incrementAndGet();

            return taskExecution;
        }
    }

    //TODO: check if cancel is successful before decreasing count
    private void cancelOnTaskThread(Task t) {
        synchronized (taskCount) {
            taskThread.cancel(t);
            taskCount.decrementAndGet();
            taskCount.notify();
        }
    }

    private void cancelOnTaskThread(List<? extends Task> tasks) {
        synchronized (taskCount) {
            taskThread.cancel(tasks);
            taskCount.addAndGet(-tasks.size());
            taskCount.notify();
        }
    }

    public static class TaskWithExecution<T extends Task> {
        private final T task;
        private final TaskExecution taskExecution;

        public TaskWithExecution(T task, TaskExecution taskExecution) {
            this.task = task;
            this.taskExecution = taskExecution;
        }

        public T getTask() {
            return task;
        }

        public TaskExecution getTaskExecution() {
            return taskExecution;
        }
    }
}
