package be.iminds.ilabt.jfed.preferences;

import be.iminds.ilabt.jfed.lowlevel.RetrySettings;
import be.iminds.ilabt.jfed.util.JFedUtils;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;

/**
 * JFedCorePreferences
 */
public class JFedCorePreferences implements JFedPreferences {
    private final static Logger LOG = LoggerFactory.getLogger(JFedCorePreferences.class);
    protected final File propertiesFile;
    protected final PropertiesConfiguration config;

    public JFedCorePreferences() {
        String jFedUserDataDir = JFedUtils.getUserDataDirectory();
         //jFedUserDataDir is null if user has no home dir and thus no stored config

        try {
            boolean fileOk;
            if (jFedUserDataDir != null) {
                File userDataDir = new File(jFedUserDataDir);
                propertiesFile = new File(userDataDir, "experimenter-ssh.properties");
                if (!propertiesFile.exists()) {
                    if (!userDataDir.exists()) {
                        LOG.debug("Will try to create " + userDataDir.getPath() + " dir");
                        userDataDir.mkdirs();

                        if (userDataDir.exists() && userDataDir.isDirectory()) {
                            LOG.info("Successfully created " + userDataDir.getPath() + " dir");
                        } else {
                            LOG.error("Could not create jFed user data dir at \"" + userDataDir.getPath() + "\". Fallback: Settings will NOT be saved!");

                            fileOk = false;
                        }
                    }

                    if (userDataDir.exists())
                        try {
                            propertiesFile.createNewFile();
                            fileOk = true;
                        } catch (IOException e) {
                            LOG.error("Could not create PropertiesConfiguration file: \"" + e.getMessage() + "\". Fallback: Settings will NOT be saved!", e);
                            //throw new RuntimeException("Could not create PropertiesConfiguration", e);

                            fileOk = false;
                        }
                    else
                        fileOk = false;
                } else {
                    fileOk = propertiesFile.canRead();
                    if (!fileOk)
                        LOG.error("WARNING: cannot read jFed properties file: \"" + propertiesFile.getPath() + "\". Fallback: Settings will NOT be loaded!");
                }
            } else {
                propertiesFile = null;
                fileOk = false;
            }

            if (fileOk)
                config = new PropertiesConfiguration(propertiesFile);
            else {
                //fallback: don't save settings to file
                config = new PropertiesConfiguration();
            }
        } catch (ConfigurationException e) {
            LOG.error("Could not create PropertiesConfiguration", e);
            throw new RuntimeException("Could not create PropertiesConfiguration", e);
        }

        //no autosave! config.setAutoSave(true);
        config.setThrowExceptionOnMissing(false);

        //use defaults to fill in blanks
        fillinBlanks(config);
    }

    public void fillinBlanks() {
        fillinBlanks(config);
    }

    protected void fillinBlanks(PropertiesConfiguration config) {
        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDS_XML_SOURCE.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDS_XML_SOURCE.key, "AUTO");

        if (!config.containsKey(CorePreferenceKey.PREF_DEBUG_CPU.key))
            config.setProperty(CorePreferenceKey.PREF_DEBUG_CPU.key, false);

        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_OVERWRITE.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_OVERWRITE.key, false);
        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_ROOT_LEVEL.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_ROOT_LEVEL.key, "INFO");
//        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_APPENDER_BUGREPORT_LEVEL.key))
//            config.setProperty(CorePreferenceKey.PREF_LOGGING_APPENDER_BUGREPORT_LEVEL.key, "INFO");
//        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_APPENDER_STDOUT_LEVEL.key))
//            config.setProperty(CorePreferenceKey.PREF_LOGGING_APPENDER_STDOUT_LEVEL.key, "INFO");
        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_EXTTERM_LEVEL.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_EXTTERM_LEVEL.key, "DEBUG");

        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM.key, "AUTO");

        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD.key, "BOTH");

        if (getRetrySettings() == null)
            setRetrySettings(new RetrySettings()/* = use internal default values*/);
    }

    public boolean containsPreference(PreferenceKey key) {
        return config.containsKey(key.getKey());
    }

    public boolean hasKey(PreferenceKey key) {
        return config.containsKey(key.getKey());
    }

    public String getString(PreferenceKey key) {
        return config.getString(key.getKey(), null);
    }

    public void setString(PreferenceKey key, String value) {
        config.setProperty(key.getKey(), value);
    }

    public Integer getInt(PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;
        return config.getInt(key.getKey());
    }

    public Integer getInt(PreferenceKey key, int defaultValue) {
        if (!config.containsKey(key.getKey())) return defaultValue;
        return config.getInt(key.getKey());
    }

    public void setInt(PreferenceKey key, Integer value) {
        config.setProperty(key.getKey(), value);
    }

    public Boolean getBoolean(PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;
        return config.getBoolean(key.getKey());
    }

    public Boolean getBoolean(PreferenceKey key, boolean defaultValue) {
        if (!config.containsKey(key.getKey())) return defaultValue;
        return config.getBoolean(key.getKey());
    }

    public void setBoolean(PreferenceKey key, Boolean value) {
        config.setProperty(key.getKey(), value);
    }

    public File getFile(PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;
        return new File(config.getString(key.getKey()));
    }

    public void setFile(PreferenceKey key, File value) {
        config.setProperty(key.getKey(), value.getPath());
    }

    public void clear(PreferenceKey key) {
        config.clearProperty(key.getKey());
    }

    public File getPropertiesFile() {
        return propertiesFile;
    }



    @Override
    public void addPreferenceChangeListener(PreferenceKey prefDebugCpu, PreferenceChangeListener preferenceChangeListener) {
        //TODO: implement
    }


//  no longer needed: AuthorityFinder always checks the current preferences
//    public void updateForceExoSM(AuthorityListModel authorityListModel) {


    @Override
    public void setRetrySettings(RetrySettings retrySettings) {
        setInt(CorePreferenceKey.RETRY_TIMEOUT_MAX, retrySettings.getTimeoutMaxRetries());
        setInt(CorePreferenceKey.RETRY_BUSY_MAX, retrySettings.getBusyMaxRetries());
        setInt(CorePreferenceKey.RETRY_POSTEMPERR_MAX, retrySettings.getPossibleTemporaryErrorMaxRetries());

        setString(CorePreferenceKey.RETRY_TIMEOUT_MS_DELAY, retrySettings.getTimeoutMsBeforeRetry()+"");
        setString(CorePreferenceKey.RETRY_BUSY_MS_DELAY, retrySettings.getBusyMsBeforeRetry()+"");
        setString(CorePreferenceKey.RETRY_POSTEMPERR_MS_DELAY, retrySettings.getPossibleTemporaryErrorMsBeforeRetry() + "");
    }

    @Override
    public RetrySettings getRetrySettings() {
        Integer timeoutMaxRetries = getInt(CorePreferenceKey.RETRY_TIMEOUT_MAX);
        Integer busyMaxRetries = getInt(CorePreferenceKey.RETRY_BUSY_MAX);
        Integer possibleTemporaryErrorMaxRetries = getInt(CorePreferenceKey.RETRY_POSTEMPERR_MAX);

        String timeoutMsBeforeRetry = getString(CorePreferenceKey.RETRY_TIMEOUT_MS_DELAY);
        String busyMsBeforeRetry = getString(CorePreferenceKey.RETRY_BUSY_MS_DELAY);
        String possibleTemporaryErrorMsBeforeRetry = getString(CorePreferenceKey.RETRY_POSTEMPERR_MS_DELAY);

        if (timeoutMaxRetries == null) return null;
        if (busyMaxRetries == null) return null;
        if (possibleTemporaryErrorMaxRetries == null) return null;
        if (timeoutMsBeforeRetry == null) return null;
        if (busyMsBeforeRetry == null) return null;
        if (possibleTemporaryErrorMsBeforeRetry == null) return null;

        RetrySettings res;
        try {
            return new RetrySettings(
                    timeoutMaxRetries, Long.parseLong(timeoutMsBeforeRetry),
                    busyMaxRetries, Long.parseLong(busyMsBeforeRetry),
                    possibleTemporaryErrorMaxRetries, Long.parseLong(possibleTemporaryErrorMsBeforeRetry)
            );
        } catch (NumberFormatException e) {
            LOG.error("Error parsing a RetrySettings", e);
            return null;
        }
    }


}
